import logging
import os

from crewai.tools import BaseTool

from crewai_tools.adapters.zapier_adapter import ZapierActionsAdapter


logger = logging.getLogger(__name__)


def ZapierActionTools(  # noqa: N802
    zapier_api_key: str | None = None, action_list: list[str] | None = None
) -> list[BaseTool]:
    """Factory function that returns Zapier action tools.

    Args:
        zapier_api_key: The API key for Zapier.
        action_list: Optional list of specific tool names to include.

    Returns:
        A list of Zapier action tools.
    """
    if zapier_api_key is None:
        zapier_api_key = os.getenv("ZAPIER_API_KEY")
        if zapier_api_key is None:
            logger.error("ZAPIER_API_KEY is not set")
            raise ValueError("ZAPIER_API_KEY is not set")
    adapter = ZapierActionsAdapter(zapier_api_key)
    all_tools = adapter.tools()

    if action_list is None:
        return all_tools

    return [tool for tool in all_tools if tool.name in action_list]



from collections.abc import Callable
from typing import Generic, TypeVar

from crewai.tools import BaseTool


T = TypeVar("T", bound=BaseTool)


class ToolCollection(list, Generic[T]):
    """A collection of tools that can be accessed by index or name.

    This class extends the built-in list to provide dictionary-like
    access to tools based on their name property.

    Usage:
        tools = ToolCollection(list_of_tools)
        # Access by index (regular list behavior)
        first_tool = tools[0]
        # Access by name (new functionality)
        search_tool = tools["search"]
    """

    def __init__(self, tools: list[T] | None = None):
        super().__init__(tools or [])
        self._name_cache: dict[str, T] = {}
        self._build_name_cache()

    def _build_name_cache(self) -> None:
        self._name_cache = {tool.name.lower(): tool for tool in self}

    def __getitem__(self, key: int | str) -> T:
        if isinstance(key, str):
            return self._name_cache[key.lower()]
        return super().__getitem__(key)

    def append(self, tool: T) -> None:
        super().append(tool)
        self._name_cache[tool.name.lower()] = tool

    def extend(self, tools: list[T]) -> None:
        super().extend(tools)
        self._build_name_cache()

    def insert(self, index: int, tool: T) -> None:
        super().insert(index, tool)
        self._name_cache[tool.name.lower()] = tool

    def remove(self, tool: T) -> None:
        super().remove(tool)
        if tool.name.lower() in self._name_cache:
            del self._name_cache[tool.name.lower()]

    def pop(self, index: int = -1) -> T:
        tool = super().pop(index)
        if tool.name.lower() in self._name_cache:
            del self._name_cache[tool.name.lower()]
        return tool

    def filter_by_names(self, names: list[str] | None = None) -> "ToolCollection[T]":
        if names is None:
            return self

        return ToolCollection(
            [
                tool
                for name in names
                if (tool := self._name_cache.get(name.lower())) is not None
            ]
        )

    def filter_where(self, func: Callable[[T], bool]) -> "ToolCollection[T]":
        return ToolCollection([tool for tool in self if func(tool)])

    def clear(self) -> None:
        super().clear()
        self._name_cache.clear()



from typing import Any

from pydantic import BaseModel, ConfigDict, Field

from .serpapi_base_tool import SerpApiBaseTool


try:
    from serpapi import HTTPError
except ImportError:
    HTTPError = Any


class SerpApiGoogleShoppingToolSchema(BaseModel):
    """Input for Google Shopping."""

    search_query: str = Field(
        ..., description="Mandatory search query you want to use to Google shopping."
    )
    location: str | None = Field(
        None, description="Location you want the search to be performed in."
    )


class SerpApiGoogleShoppingTool(SerpApiBaseTool):
    model_config = ConfigDict(
        arbitrary_types_allowed=True, validate_assignment=True, frozen=False
    )
    name: str = "Google Shopping"
    description: str = (
        "A tool to perform search on Google shopping with a search_query."
    )
    args_schema: type[BaseModel] = SerpApiGoogleShoppingToolSchema

    def _run(
        self,
        **kwargs: Any,
    ) -> Any:
        try:
            results = self.client.search(
                {
                    "engine": "google_shopping",
                    "q": kwargs.get("search_query"),
                    "location": kwargs.get("location"),
                }
            ).as_dict()

            self._omit_fields(
                results,
                [
                    r"search_metadata",
                    r"search_parameters",
                    r"serpapi_.+",
                    r"filters",
                    r"pagination",
                ],
            )

            return results
        except HTTPError as e:
            return f"An error occurred: {e!s}. Some parameters may be invalid."



import csv
from io import StringIO

from crewai_tools.rag.base_loader import BaseLoader, LoaderResult
from crewai_tools.rag.source_content import SourceContent


class CSVLoader(BaseLoader):
    def load(self, source_content: SourceContent, **kwargs) -> LoaderResult:
        source_ref = source_content.source_ref

        content_str = source_content.source
        if source_content.is_url():
            content_str = self._load_from_url(content_str, kwargs)
        elif source_content.path_exists():
            content_str = self._load_from_file(content_str)

        return self._parse_csv(content_str, source_ref)

    def _load_from_url(self, url: str, kwargs: dict) -> str:
        import requests

        headers = kwargs.get(
            "headers",
            {
                "Accept": "text/csv, application/csv, text/plain",
                "User-Agent": "Mozilla/5.0 (compatible; crewai-tools CSVLoader)",
            },
        )

        try:
            response = requests.get(url, headers=headers, timeout=30)
            response.raise_for_status()
            return response.text
        except Exception as e:
            raise ValueError(f"Error fetching CSV from URL {url}: {e!s}") from e

    def _load_from_file(self, path: str) -> str:
        with open(path, "r", encoding="utf-8") as file:
            return file.read()

    def _parse_csv(self, content: str, source_ref: str) -> LoaderResult:
        try:
            csv_reader = csv.DictReader(StringIO(content))

            text_parts = []
            headers = csv_reader.fieldnames

            if headers:
                text_parts.append("Headers: " + " | ".join(headers))
                text_parts.append("-" * 50)

                for row_num, row in enumerate(csv_reader, 1):
                    row_text = " | ".join([f"{k}: {v}" for k, v in row.items() if v])
                    text_parts.append(f"Row {row_num}: {row_text}")

            text = "\n".join(text_parts)

            metadata = {
                "format": "csv",
                "columns": headers,
                "rows": len(text_parts) - 2 if headers else 0,
            }

        except Exception as e:
            text = content
            metadata = {"format": "csv", "parse_error": str(e)}

        return LoaderResult(
            content=text,
            source=source_ref,
            metadata=metadata,
            doc_id=self.generate_doc_id(source_ref=source_ref, content=text),
        )



import os
import tempfile

from crewai_tools.rag.base_loader import BaseLoader, LoaderResult
from crewai_tools.rag.source_content import SourceContent


class DOCXLoader(BaseLoader):
    def load(self, source_content: SourceContent, **kwargs) -> LoaderResult:
        try:
            from docx import Document as DocxDocument
        except ImportError as e:
            raise ImportError(
                "python-docx is required for DOCX loading. Install with: 'uv pip install python-docx' or pip install crewai-tools[rag]"
            ) from e

        source_ref = source_content.source_ref

        if source_content.is_url():
            temp_file = self._download_from_url(source_ref, kwargs)
            try:
                return self._load_from_file(temp_file, source_ref, DocxDocument)
            finally:
                os.unlink(temp_file)
        elif source_content.path_exists():
            return self._load_from_file(source_ref, source_ref, DocxDocument)
        else:
            raise ValueError(
                f"Source must be a valid file path or URL, got: {source_content.source}"
            )

    def _download_from_url(self, url: str, kwargs: dict) -> str:
        import requests

        headers = kwargs.get(
            "headers",
            {
                "Accept": "application/vnd.openxmlformats-officedocument.wordprocessingml.document",
                "User-Agent": "Mozilla/5.0 (compatible; crewai-tools DOCXLoader)",
            },
        )

        try:
            response = requests.get(url, headers=headers, timeout=30)
            response.raise_for_status()

            # Create temporary file to save the DOCX content
            with tempfile.NamedTemporaryFile(suffix=".docx", delete=False) as temp_file:
                temp_file.write(response.content)
                return temp_file.name
        except Exception as e:
            raise ValueError(f"Error fetching DOCX from URL {url}: {e!s}") from e

    def _load_from_file(
        self,
        file_path: str,
        source_ref: str,
        DocxDocument,  # noqa: N803
    ) -> LoaderResult:
        try:
            doc = DocxDocument(file_path)

            text_parts = []
            for paragraph in doc.paragraphs:
                if paragraph.text.strip():
                    text_parts.append(paragraph.text)  # noqa: PERF401

            content = "\n".join(text_parts)

            metadata = {
                "format": "docx",
                "paragraphs": len(doc.paragraphs),
                "tables": len(doc.tables),
            }

            return LoaderResult(
                content=content,
                source=source_ref,
                metadata=metadata,
                doc_id=self.generate_doc_id(source_ref=source_ref, content=content),
            )

        except Exception as e:
            raise ValueError(f"Error loading DOCX file: {e!s}") from e



"""Optical Character Recognition (OCR) Tool.

This tool provides functionality for extracting text from images using supported LLMs. Make sure your model supports the `vision` feature.
"""

import base64

from crewai import LLM
from crewai.tools.base_tool import BaseTool
from pydantic import BaseModel, PrivateAttr


class OCRToolSchema(BaseModel):
    """Input schema for Optical Character Recognition Tool.

    Attributes:
        image_path_url (str): Path to a local image file or URL of an image.
            For local files, provide the absolute or relative path.
            For remote images, provide the complete URL starting with 'http' or 'https'.
    """

    image_path_url: str = "The image path or URL."


class OCRTool(BaseTool):
    """A tool for performing Optical Character Recognition on images.

    This tool leverages LLMs to extract text from images. It can process
    both local image files and images available via URLs.

    Attributes:
        name (str): Name of the tool.
        description (str): Description of the tool's functionality.
        args_schema (Type[BaseModel]): Pydantic schema for input validation.

    Private Attributes:
        _llm (Optional[LLM]): Language model instance for making API calls.
    """

    name: str = "Optical Character Recognition Tool"
    description: str = "This tool uses an LLM's API to extract text from an image file."
    _llm: LLM | None = PrivateAttr(default=None)

    args_schema: type[BaseModel] = OCRToolSchema

    def __init__(self, llm: LLM = None, **kwargs):
        """Initialize the OCR tool.

        Args:
            llm (LLM, optional): Language model instance to use for API calls.
                If not provided, a default LLM with gpt-4o model will be used.
            **kwargs: Additional arguments passed to the parent class.
        """
        super().__init__(**kwargs)

        if llm is None:
            # Use the default LLM
            llm = LLM(
                model="gpt-4o",
                temperature=0.7,
            )

        self._llm = llm

    def _run(self, **kwargs) -> str:
        """Execute the OCR operation on the provided image.

        Args:
            **kwargs: Keyword arguments containing the image_path_url.

        Returns:
            str: Extracted text from the image.
                If no image path/URL is provided, returns an error message.

        Note:
            The method handles both local image files and remote URLs:
            - For local files: The image is read and encoded to base64
            - For URLs: The URL is passed directly to the Vision API
        """
        image_path_url = kwargs.get("image_path_url")

        if not image_path_url:
            return "Image Path or URL is required."

        if image_path_url.startswith("http"):
            image_data = image_path_url
        else:
            base64_image = self._encode_image(image_path_url)
            image_data = f"data:image/jpeg;base64,{base64_image}"

        messages = [
            {
                "role": "system",
                "content": "You are an expert OCR specialist. Extract complete text from the provided image. Provide the result as a raw text.",
            },
            {
                "role": "user",
                "content": [
                    {
                        "type": "image_url",
                        "image_url": {"url": image_data},
                    }
                ],
            },
        ]

        return self._llm.call(messages=messages)

    def _encode_image(self, image_path: str):
        """Encode an image file to base64 format.

        Args:
            image_path (str): Path to the local image file.

        Returns:
            str: Base64-encoded image data as a UTF-8 string.
        """
        with open(image_path, "rb") as image_file:
            return base64.b64encode(image_file.read()).decode("utf-8")



from unittest.mock import Mock, patch

from crewai_tools.tools.brightdata_tool.brightdata_unlocker import (
    BrightDataWebUnlockerTool,
)
import requests


@patch.dict(
    "os.environ",
    {"BRIGHT_DATA_API_KEY": "test_api_key", "BRIGHT_DATA_ZONE": "test_zone"},
)
@patch("crewai_tools.tools.brightdata_tool.brightdata_unlocker.requests.post")
def test_run_success_html(mock_post):
    mock_response = Mock()
    mock_response.status_code = 200
    mock_response.text = "<html><body>Test</body></html>"
    mock_response.raise_for_status = Mock()
    mock_post.return_value = mock_response

    tool = BrightDataWebUnlockerTool()
    tool._run(url="https://example.com", format="html", save_file=False)


@patch.dict(
    "os.environ",
    {"BRIGHT_DATA_API_KEY": "test_api_key", "BRIGHT_DATA_ZONE": "test_zone"},
)
@patch("crewai_tools.tools.brightdata_tool.brightdata_unlocker.requests.post")
def test_run_success_json(mock_post):
    mock_response = Mock()
    mock_response.status_code = 200
    mock_response.text = "mock response text"
    mock_response.raise_for_status = Mock()
    mock_post.return_value = mock_response

    tool = BrightDataWebUnlockerTool()
    result = tool._run(url="https://example.com", format="json")

    assert isinstance(result, str)


@patch.dict(
    "os.environ",
    {"BRIGHT_DATA_API_KEY": "test_api_key", "BRIGHT_DATA_ZONE": "test_zone"},
)
@patch("crewai_tools.tools.brightdata_tool.brightdata_unlocker.requests.post")
def test_run_http_error(mock_post):
    mock_response = Mock()
    mock_response.status_code = 403
    mock_response.text = "Forbidden"
    mock_response.raise_for_status.side_effect = requests.HTTPError(
        response=mock_response
    )
    mock_post.return_value = mock_response

    tool = BrightDataWebUnlockerTool()
    result = tool._run(url="https://example.com")

    assert "HTTP Error" in result
    assert "Forbidden" in result



"""MCPServer for CrewAI."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from crewai.tools import BaseTool

from crewai_tools.adapters.tool_collection import ToolCollection


logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from mcp import StdioServerParameters
    from mcpadapt.core import MCPAdapt
    from mcpadapt.crewai_adapter import CrewAIAdapter


try:
    from mcp import StdioServerParameters
    from mcpadapt.core import MCPAdapt
    from mcpadapt.crewai_adapter import CrewAIAdapter

    MCP_AVAILABLE = True
except ImportError:
    MCP_AVAILABLE = False


class MCPServerAdapter:
    """Manages the lifecycle of an MCP server and make its tools available to CrewAI.

    Note: tools can only be accessed after the server has been started with the
        `start()` method.

    Attributes:
        tools: The CrewAI tools available from the MCP server.

    Usage:
        # context manager + stdio
        with MCPServerAdapter(...) as tools:
            # tools is now available

        # context manager + sse
        with MCPServerAdapter({"url": "http://localhost:8000/sse"}) as tools:
            # tools is now available

        # context manager with filtered tools
        with MCPServerAdapter(..., "tool1", "tool2") as filtered_tools:
            # only tool1 and tool2 are available

        # context manager with custom connect timeout (60 seconds)
        with MCPServerAdapter(..., connect_timeout=60) as tools:
            # tools is now available with longer timeout

        # manually stop mcp server
        try:
            mcp_server = MCPServerAdapter(...)
            tools = mcp_server.tools  # all tools

            # or with filtered tools and custom timeout
            mcp_server = MCPServerAdapter(..., "tool1", "tool2", connect_timeout=45)
            filtered_tools = mcp_server.tools  # only tool1 and tool2
            ...
        finally:
            mcp_server.stop()

        # Best practice is ensure cleanup is done after use.
        mcp_server.stop() # run after crew().kickoff()
    """

    def __init__(
        self,
        serverparams: StdioServerParameters | dict[str, Any],
        *tool_names: str,
        connect_timeout: int = 30,
    ) -> None:
        """Initialize the MCP Server.

        Args:
            serverparams: The parameters for the MCP server it supports either a
                `StdioServerParameters` or a `dict` respectively for STDIO and SSE.
            *tool_names: Optional names of tools to filter. If provided, only tools with
                matching names will be available.
            connect_timeout: Connection timeout in seconds to the MCP server (default is 30s).

        """
        super().__init__()
        self._adapter = None
        self._tools = None
        self._tool_names = list(tool_names) if tool_names else None

        if not MCP_AVAILABLE:
            import click

            if click.confirm(
                "You are missing the 'mcp' package. Would you like to install it?"
            ):
                import subprocess

                try:
                    subprocess.run(["uv", "add", "mcp crewai-tools[mcp]"], check=True)  # noqa: S607

                except subprocess.CalledProcessError as e:
                    raise ImportError("Failed to install mcp package") from e
            else:
                raise ImportError(
                    "`mcp` package not found, please run `uv add crewai-tools[mcp]`"
                )

        try:
            self._serverparams = serverparams
            self._adapter = MCPAdapt(
                self._serverparams, CrewAIAdapter(), connect_timeout
            )
            self.start()

        except Exception as e:
            if self._adapter is not None:
                try:
                    self.stop()
                except Exception as stop_e:
                    logger.error(f"Error during stop cleanup: {stop_e}")
            raise RuntimeError(f"Failed to initialize MCP Adapter: {e}") from e

    def start(self):
        """Start the MCP server and initialize the tools."""
        self._tools = self._adapter.__enter__()

    def stop(self):
        """Stop the MCP server."""
        self._adapter.__exit__(None, None, None)

    @property
    def tools(self) -> ToolCollection[BaseTool]:
        """The CrewAI tools available from the MCP server.

        Raises:
            ValueError: If the MCP server is not started.

        Returns:
            The CrewAI tools available from the MCP server.
        """
        if self._tools is None:
            raise ValueError(
                "MCP server not started, run `mcp_server.start()` first before accessing `tools`"
            )

        tools_collection = ToolCollection(self._tools)
        if self._tool_names:
            return tools_collection.filter_by_names(self._tool_names)
        return tools_collection

    def __enter__(self):
        """Enter the context manager. Note that `__init__()` already starts the MCP server.
        So tools should already be available.
        """
        return self.tools

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit the context manager."""
        return self._adapter.__exit__(exc_type, exc_value, traceback)



import datetime
import json
import logging
import os
from typing import Any

from crewai.tools import BaseTool, EnvVar
from pydantic import BaseModel, Field
import requests


logger = logging.getLogger(__name__)


def _save_results_to_file(content: str) -> None:
    """Saves the search results to a file."""
    try:
        filename = f"search_results_{datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}.txt"
        with open(filename, "w") as file:
            file.write(content)
        logger.info(f"Results saved to {filename}")
    except IOError as e:
        logger.error(f"Failed to save results to file: {e}")
        raise


class SerperDevToolSchema(BaseModel):
    """Input for SerperDevTool."""

    search_query: str = Field(
        ..., description="Mandatory search query you want to use to search the internet"
    )


class SerperDevTool(BaseTool):
    name: str = "Search the internet with Serper"
    description: str = (
        "A tool that can be used to search the internet with a search_query. "
        "Supports different search types: 'search' (default), 'news'"
    )
    args_schema: type[BaseModel] = SerperDevToolSchema
    base_url: str = "https://google.serper.dev"
    n_results: int = 10
    save_file: bool = False
    search_type: str = "search"
    country: str | None = ""
    location: str | None = ""
    locale: str | None = ""
    env_vars: list[EnvVar] = Field(
        default_factory=lambda: [
            EnvVar(
                name="SERPER_API_KEY", description="API key for Serper", required=True
            ),
        ]
    )

    def _get_search_url(self, search_type: str) -> str:
        """Get the appropriate endpoint URL based on search type."""
        search_type = search_type.lower()
        allowed_search_types = ["search", "news"]
        if search_type not in allowed_search_types:
            raise ValueError(
                f"Invalid search type: {search_type}. Must be one of: {', '.join(allowed_search_types)}"
            )
        return f"{self.base_url}/{search_type}"

    def _process_knowledge_graph(self, kg: dict) -> dict:
        """Process knowledge graph data from search results."""
        return {
            "title": kg.get("title", ""),
            "type": kg.get("type", ""),
            "website": kg.get("website", ""),
            "imageUrl": kg.get("imageUrl", ""),
            "description": kg.get("description", ""),
            "descriptionSource": kg.get("descriptionSource", ""),
            "descriptionLink": kg.get("descriptionLink", ""),
            "attributes": kg.get("attributes", {}),
        }

    def _process_organic_results(self, organic_results: list) -> list:
        """Process organic search results."""
        processed_results = []
        for result in organic_results[: self.n_results]:
            try:
                result_data = {
                    "title": result["title"],
                    "link": result["link"],
                    "snippet": result.get("snippet", ""),
                    "position": result.get("position"),
                }

                if "sitelinks" in result:
                    result_data["sitelinks"] = [
                        {
                            "title": sitelink.get("title", ""),
                            "link": sitelink.get("link", ""),
                        }
                        for sitelink in result["sitelinks"]
                    ]

                processed_results.append(result_data)
            except KeyError:  # noqa: PERF203
                logger.warning(f"Skipping malformed organic result: {result}")
                continue
        return processed_results

    def _process_people_also_ask(self, paa_results: list) -> list:
        """Process 'People Also Ask' results."""
        processed_results = []
        for result in paa_results[: self.n_results]:
            try:
                result_data = {
                    "question": result["question"],
                    "snippet": result.get("snippet", ""),
                    "title": result.get("title", ""),
                    "link": result.get("link", ""),
                }
                processed_results.append(result_data)
            except KeyError:  # noqa: PERF203
                logger.warning(f"Skipping malformed PAA result: {result}")
                continue
        return processed_results

    def _process_related_searches(self, related_results: list) -> list:
        """Process related search results."""
        processed_results = []
        for result in related_results[: self.n_results]:
            try:
                processed_results.append({"query": result["query"]})
            except KeyError:  # noqa: PERF203
                logger.warning(f"Skipping malformed related search result: {result}")
                continue
        return processed_results

    def _process_news_results(self, news_results: list) -> list:
        """Process news search results."""
        processed_results = []
        for result in news_results[: self.n_results]:
            try:
                result_data = {
                    "title": result["title"],
                    "link": result["link"],
                    "snippet": result.get("snippet", ""),
                    "date": result.get("date", ""),
                    "source": result.get("source", ""),
                    "imageUrl": result.get("imageUrl", ""),
                }
                processed_results.append(result_data)
            except KeyError:  # noqa: PERF203
                logger.warning(f"Skipping malformed news result: {result}")
                continue
        return processed_results

    def _make_api_request(self, search_query: str, search_type: str) -> dict:
        """Make API request to Serper."""
        search_url = self._get_search_url(search_type)
        payload = {"q": search_query, "num": self.n_results}

        if self.country != "":
            payload["gl"] = self.country
        if self.location != "":
            payload["location"] = self.location
        if self.locale != "":
            payload["hl"] = self.locale

        headers = {
            "X-API-KEY": os.environ["SERPER_API_KEY"],
            "content-type": "application/json",
        }
        payload = json.dumps(payload)

        response = None
        try:
            response = requests.post(
                search_url, headers=headers, json=json.loads(payload), timeout=10
            )
            response.raise_for_status()
            results = response.json()
            if not results:
                logger.error("Empty response from Serper API")
                raise ValueError("Empty response from Serper API")
            return results
        except requests.exceptions.RequestException as e:
            error_msg = f"Error making request to Serper API: {e}"
            if response is not None and hasattr(response, "content"):
                error_msg += f"\nResponse content: {response.content}"
            logger.error(error_msg)
            raise
        except json.JSONDecodeError as e:
            if response is not None and hasattr(response, "content"):
                logger.error(f"Error decoding JSON response: {e}")
                logger.error(f"Response content: {response.content}")
            else:
                logger.error(
                    f"Error decoding JSON response: {e} (No response content available)"
                )
            raise

    def _process_search_results(self, results: dict, search_type: str) -> dict:
        """Process search results based on search type."""
        formatted_results = {}

        if search_type == "search":
            if "knowledgeGraph" in results:
                formatted_results["knowledgeGraph"] = self._process_knowledge_graph(
                    results["knowledgeGraph"]
                )

            if "organic" in results:
                formatted_results["organic"] = self._process_organic_results(
                    results["organic"]
                )

            if "peopleAlsoAsk" in results:
                formatted_results["peopleAlsoAsk"] = self._process_people_also_ask(
                    results["peopleAlsoAsk"]
                )

            if "relatedSearches" in results:
                formatted_results["relatedSearches"] = self._process_related_searches(
                    results["relatedSearches"]
                )

        elif search_type == "news":
            if "news" in results:
                formatted_results["news"] = self._process_news_results(results["news"])

        return formatted_results

    def _run(self, **kwargs: Any) -> Any:
        """Execute the search operation."""
        search_query = kwargs.get("search_query") or kwargs.get("query")
        search_type = kwargs.get("search_type", self.search_type)
        save_file = kwargs.get("save_file", self.save_file)

        results = self._make_api_request(search_query, search_type)

        formatted_results = {
            "searchParameters": {
                "q": search_query,
                "type": search_type,
                **results.get("searchParameters", {}),
            }
        }

        formatted_results.update(self._process_search_results(results, search_type))
        formatted_results["credits"] = results.get("credits", 1)

        if save_file:
            _save_results_to_file(json.dumps(formatted_results, indent=2))

        return formatted_results



from .agents.invoke_agent_tool import BedrockInvokeAgentTool
from .browser import create_browser_toolkit
from .code_interpreter import create_code_interpreter_toolkit
from .knowledge_base.retriever_tool import BedrockKBRetrieverTool


__all__ = [
    "BedrockInvokeAgentTool",
    "BedrockKBRetrieverTool",
    "create_browser_toolkit",
    "create_code_interpreter_toolkit",
]



"""Crewai Enterprise Tools."""

import json
import re
from typing import Any, Optional, Union, cast, get_origin

from crewai.tools import BaseTool
from pydantic import Field, create_model
import requests

from crewai_tools.tools.crewai_platform_tools.misc import (
    get_platform_api_base_url,
    get_platform_integration_token,
)


class AllOfSchemaAnalyzer:
    """Helper class to analyze and merge allOf schemas."""

    def __init__(self, schemas: list[dict[str, Any]]):
        self.schemas = schemas
        self._explicit_types: list[str] = []
        self._merged_properties: dict[str, Any] = {}
        self._merged_required: list[str] = []
        self._analyze_schemas()

    def _analyze_schemas(self) -> None:
        """Analyze all schemas and extract relevant information."""
        for schema in self.schemas:
            if "type" in schema:
                self._explicit_types.append(schema["type"])

            # Merge object properties
            if schema.get("type") == "object" and "properties" in schema:
                self._merged_properties.update(schema["properties"])
                if "required" in schema:
                    self._merged_required.extend(schema["required"])

    def has_consistent_type(self) -> bool:
        """Check if all schemas have the same explicit type."""
        return len(set(self._explicit_types)) == 1 if self._explicit_types else False

    def get_consistent_type(self) -> type[Any]:
        """Get the consistent type if all schemas agree."""
        if not self.has_consistent_type():
            raise ValueError("No consistent type found")

        type_mapping = {
            "string": str,
            "integer": int,
            "number": float,
            "boolean": bool,
            "array": list,
            "object": dict,
            "null": type(None),
        }
        return type_mapping.get(self._explicit_types[0], str)

    def has_object_schemas(self) -> bool:
        """Check if any schemas are object types with properties."""
        return bool(self._merged_properties)

    def get_merged_properties(self) -> dict[str, Any]:
        """Get merged properties from all object schemas."""
        return self._merged_properties

    def get_merged_required_fields(self) -> list[str]:
        """Get merged required fields from all object schemas."""
        return list(set(self._merged_required))  # Remove duplicates

    def get_fallback_type(self) -> type[Any]:
        """Get a fallback type when merging fails."""
        if self._explicit_types:
            # Use the first explicit type
            type_mapping = {
                "string": str,
                "integer": int,
                "number": float,
                "boolean": bool,
                "array": list,
                "object": dict,
                "null": type(None),
            }
            return type_mapping.get(self._explicit_types[0], str)
        return str


class CrewAIPlatformActionTool(BaseTool):
    action_name: str = Field(default="", description="The name of the action")
    action_schema: dict[str, Any] = Field(
        default_factory=dict, description="The schema of the action"
    )

    def __init__(
        self,
        description: str,
        action_name: str,
        action_schema: dict[str, Any],
    ):
        self._model_registry: dict[str, type[Any]] = {}
        self._base_name = self._sanitize_name(action_name)

        schema_props, required = self._extract_schema_info(action_schema)

        field_definitions: dict[str, Any] = {}
        for param_name, param_details in schema_props.items():
            param_desc = param_details.get("description", "")
            is_required = param_name in required

            try:
                field_type = self._process_schema_type(
                    param_details, self._sanitize_name(param_name).title()
                )
            except Exception:
                field_type = str

            field_definitions[param_name] = self._create_field_definition(
                field_type, is_required, param_desc
            )

        if field_definitions:
            try:
                args_schema = create_model(
                    f"{self._base_name}Schema", **field_definitions
                )
            except Exception:
                args_schema = create_model(
                    f"{self._base_name}Schema",
                    input_text=(str, Field(description="Input for the action")),
                )
        else:
            args_schema = create_model(
                f"{self._base_name}Schema",
                input_text=(str, Field(description="Input for the action")),
            )

        super().__init__(
            name=action_name.lower().replace(" ", "_"),
            description=description,
            args_schema=args_schema,
        )
        self.action_name = action_name
        self.action_schema = action_schema

    @staticmethod
    def _sanitize_name(name: str) -> str:
        name = name.lower().replace(" ", "_")
        sanitized = re.sub(r"[^a-zA-Z0-9_]", "", name)
        parts = sanitized.split("_")
        return "".join(word.capitalize() for word in parts if word)

    @staticmethod
    def _extract_schema_info(
        action_schema: dict[str, Any],
    ) -> tuple[dict[str, Any], list[str]]:
        schema_props = (
            action_schema.get("function", {})
            .get("parameters", {})
            .get("properties", {})
        )
        required = (
            action_schema.get("function", {}).get("parameters", {}).get("required", [])
        )
        return schema_props, required

    def _process_schema_type(self, schema: dict[str, Any], type_name: str) -> type[Any]:
        """
        Process a JSON Schema type definition into a Python type.

        Handles complex schema constructs like anyOf, oneOf, allOf, enums, arrays, and objects.
        """
        # Handle composite schema types (anyOf, oneOf, allOf)
        if composite_type := self._process_composite_schema(schema, type_name):
            return composite_type

        # Handle primitive types and simple constructs
        return self._process_primitive_schema(schema, type_name)

    def _process_composite_schema(
        self, schema: dict[str, Any], type_name: str
    ) -> type[Any] | None:
        """Process composite schema types: anyOf, oneOf, allOf."""
        if "anyOf" in schema:
            return self._process_any_of_schema(schema["anyOf"], type_name)
        if "oneOf" in schema:
            return self._process_one_of_schema(schema["oneOf"], type_name)
        if "allOf" in schema:
            return self._process_all_of_schema(schema["allOf"], type_name)
        return None

    def _process_any_of_schema(
        self, any_of_types: list[dict[str, Any]], type_name: str
    ) -> type[Any]:
        """Process anyOf schema - creates Union of possible types."""
        is_nullable = any(t.get("type") == "null" for t in any_of_types)
        non_null_types = [t for t in any_of_types if t.get("type") != "null"]

        if not non_null_types:
            return cast(
                type[Any], cast(object, str | None)
            )  # fallback for only-null case

        base_type = (
            self._process_schema_type(non_null_types[0], type_name)
            if len(non_null_types) == 1
            else self._create_union_type(non_null_types, type_name, "AnyOf")
        )
        return base_type | None if is_nullable else base_type  # type: ignore[return-value]

    def _process_one_of_schema(
        self, one_of_types: list[dict[str, Any]], type_name: str
    ) -> type[Any]:
        """Process oneOf schema - creates Union of mutually exclusive types."""
        return (
            self._process_schema_type(one_of_types[0], type_name)
            if len(one_of_types) == 1
            else self._create_union_type(one_of_types, type_name, "OneOf")
        )

    def _process_all_of_schema(
        self, all_of_schemas: list[dict[str, Any]], type_name: str
    ) -> type[Any]:
        """Process allOf schema - merges schemas that must all be satisfied."""
        if len(all_of_schemas) == 1:
            return self._process_schema_type(all_of_schemas[0], type_name)
        return self._merge_all_of_schemas(all_of_schemas, type_name)

    def _create_union_type(
        self, schemas: list[dict[str, Any]], type_name: str, prefix: str
    ) -> type[Any]:
        """Create a Union type from multiple schemas."""
        return Union[  # type: ignore  # noqa: UP007
            tuple(
                self._process_schema_type(schema, f"{type_name}{prefix}{i}")
                for i, schema in enumerate(schemas)
            )
        ]

    def _process_primitive_schema(
        self, schema: dict[str, Any], type_name: str
    ) -> type[Any]:
        """Process primitive schema types: string, number, array, object, etc."""
        json_type = schema.get("type", "string")

        if "enum" in schema:
            return self._process_enum_schema(schema, json_type)

        if json_type == "array":
            return self._process_array_schema(schema, type_name)

        if json_type == "object":
            return self._create_nested_model(schema, type_name)

        return self._map_json_type_to_python(json_type)

    def _process_enum_schema(self, schema: dict[str, Any], json_type: str) -> type[Any]:
        """Process enum schema - currently falls back to base type."""
        enum_values = schema["enum"]
        if not enum_values:
            return self._map_json_type_to_python(json_type)

        # For Literal types, we need to pass the values directly, not as a tuple
        # This is a workaround since we can't dynamically create Literal types easily
        # Fall back to the base JSON type for now
        return self._map_json_type_to_python(json_type)

    def _process_array_schema(
        self, schema: dict[str, Any], type_name: str
    ) -> type[Any]:
        items_schema = schema.get("items", {"type": "string"})
        item_type = self._process_schema_type(items_schema, f"{type_name}Item")
        return list[item_type]  # type: ignore

    def _merge_all_of_schemas(
        self, schemas: list[dict[str, Any]], type_name: str
    ) -> type[Any]:
        schema_analyzer = AllOfSchemaAnalyzer(schemas)

        if schema_analyzer.has_consistent_type():
            return schema_analyzer.get_consistent_type()

        if schema_analyzer.has_object_schemas():
            return self._create_merged_object_model(
                schema_analyzer.get_merged_properties(),
                schema_analyzer.get_merged_required_fields(),
                type_name,
            )

        return schema_analyzer.get_fallback_type()

    def _create_merged_object_model(
        self, properties: dict[str, Any], required: list[str], model_name: str
    ) -> type[Any]:
        full_model_name = f"{self._base_name}{model_name}AllOf"

        if full_model_name in self._model_registry:
            return self._model_registry[full_model_name]

        if not properties:
            return dict

        field_definitions = self._build_field_definitions(
            properties, required, model_name
        )

        try:
            merged_model = create_model(full_model_name, **field_definitions)
            self._model_registry[full_model_name] = merged_model
            return merged_model
        except Exception:
            return dict

    def _build_field_definitions(
        self, properties: dict[str, Any], required: list[str], model_name: str
    ) -> dict[str, Any]:
        field_definitions = {}

        for prop_name, prop_schema in properties.items():
            prop_desc = prop_schema.get("description", "")
            is_required = prop_name in required

            try:
                prop_type = self._process_schema_type(
                    prop_schema, f"{model_name}{self._sanitize_name(prop_name).title()}"
                )
            except Exception:
                prop_type = str

            field_definitions[prop_name] = self._create_field_definition(
                prop_type, is_required, prop_desc
            )

        return field_definitions

    def _create_nested_model(
        self, schema: dict[str, Any], model_name: str
    ) -> type[Any]:
        full_model_name = f"{self._base_name}{model_name}"

        if full_model_name in self._model_registry:
            return self._model_registry[full_model_name]

        properties = schema.get("properties", {})
        required_fields = schema.get("required", [])

        if not properties:
            return dict

        field_definitions = {}
        for prop_name, prop_schema in properties.items():
            prop_desc = prop_schema.get("description", "")
            is_required = prop_name in required_fields

            try:
                prop_type = self._process_schema_type(
                    prop_schema, f"{model_name}{self._sanitize_name(prop_name).title()}"
                )
            except Exception:
                prop_type = str

            field_definitions[prop_name] = self._create_field_definition(
                prop_type, is_required, prop_desc
            )

        try:
            nested_model = create_model(full_model_name, **field_definitions)  # type: ignore
            self._model_registry[full_model_name] = nested_model
            return nested_model
        except Exception:
            return dict

    def _create_field_definition(
        self, field_type: type[Any], is_required: bool, description: str
    ) -> tuple:
        if is_required:
            return (field_type, Field(description=description))
        if get_origin(field_type) is Union:
            return (field_type, Field(default=None, description=description))
        return (
            Optional[field_type],  # noqa: UP045
            Field(default=None, description=description),
        )

    def _map_json_type_to_python(self, json_type: str) -> type[Any]:
        type_mapping = {
            "string": str,
            "integer": int,
            "number": float,
            "boolean": bool,
            "array": list,
            "object": dict,
            "null": type(None),
        }
        return type_mapping.get(json_type, str)

    def _get_required_nullable_fields(self) -> list[str]:
        schema_props, required = self._extract_schema_info(self.action_schema)

        required_nullable_fields = []
        for param_name in required:
            param_details = schema_props.get(param_name, {})
            if self._is_nullable_type(param_details):
                required_nullable_fields.append(param_name)

        return required_nullable_fields

    def _is_nullable_type(self, schema: dict[str, Any]) -> bool:
        if "anyOf" in schema:
            return any(t.get("type") == "null" for t in schema["anyOf"])
        return schema.get("type") == "null"

    def _run(self, **kwargs) -> str:
        try:
            cleaned_kwargs = {
                key: value for key, value in kwargs.items() if value is not None
            }

            required_nullable_fields = self._get_required_nullable_fields()

            for field_name in required_nullable_fields:
                if field_name not in cleaned_kwargs:
                    cleaned_kwargs[field_name] = None

            api_url = (
                f"{get_platform_api_base_url()}/actions/{self.action_name}/execute"
            )
            token = get_platform_integration_token()
            headers = {
                "Authorization": f"Bearer {token}",
                "Content-Type": "application/json",
            }
            payload = cleaned_kwargs

            response = requests.post(
                url=api_url, headers=headers, json=payload, timeout=60
            )

            data = response.json()
            if not response.ok:
                error_message = data.get("error", {}).get("message", json.dumps(data))
                return f"API request failed: {error_message}"

            return json.dumps(data, indent=2)

        except Exception as e:
            return f"Error executing action {self.action_name}: {e!s}"



import os

from crewai import Agent, Crew, Task
from multion_tool import MultiOnTool


os.environ["OPENAI_API_KEY"] = "Your Key"

multion_browse_tool = MultiOnTool(api_key="Your Key")

# Create a new agent
Browser = Agent(
    role="Browser Agent",
    goal="control web browsers using natural language ",
    backstory="An expert browsing agent.",
    tools=[multion_browse_tool],
    verbose=True,
)

# Define tasks
browse = Task(
    description="Summarize the top 3 trending AI News headlines",
    expected_output="A summary of the top 3 trending AI News headlines",
    agent=Browser,
)


crew = Crew(agents=[Browser], tasks=[browse])

crew.kickoff()



import re

from bs4 import BeautifulSoup
import requests

from crewai_tools.rag.base_loader import BaseLoader, LoaderResult
from crewai_tools.rag.source_content import SourceContent


class WebPageLoader(BaseLoader):
    def load(self, source_content: SourceContent, **kwargs) -> LoaderResult:
        url = source_content.source
        headers = kwargs.get(
            "headers",
            {
                "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/96.0.4664.110 Safari/537.36",
                "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9",
                "Accept-Language": "en-US,en;q=0.9",
            },
        )

        try:
            response = requests.get(url, timeout=15, headers=headers)
            response.encoding = response.apparent_encoding

            soup = BeautifulSoup(response.text, "html.parser")

            for script in soup(["script", "style"]):
                script.decompose()

            text = soup.get_text(" ")
            text = re.sub("[ \t]+", " ", text)
            text = re.sub("\\s+\n\\s+", "\n", text)
            text = text.strip()

            title = (
                soup.title.string.strip() if soup.title and soup.title.string else ""
            )
            metadata = {
                "url": url,
                "title": title,
                "status_code": response.status_code,
                "content_type": response.headers.get("content-type", ""),
            }

            return LoaderResult(
                content=text,
                source=url,
                metadata=metadata,
                doc_id=self.generate_doc_id(source_ref=url, content=text),
            )

        except Exception as e:
            raise ValueError(f"Error loading webpage {url}: {e!s}") from e



import re


class RecursiveCharacterTextSplitter:
    """A text splitter that recursively splits text based on a hierarchy of separators."""

    def __init__(
        self,
        chunk_size: int = 4000,
        chunk_overlap: int = 200,
        separators: list[str] | None = None,
        keep_separator: bool = True,
    ):
        """Initialize the RecursiveCharacterTextSplitter.

        Args:
            chunk_size: Maximum size of each chunk
            chunk_overlap: Number of characters to overlap between chunks
            separators: List of separators to use for splitting (in order of preference)
            keep_separator: Whether to keep the separator in the split text
        """
        if chunk_overlap >= chunk_size:
            raise ValueError(
                f"Chunk overlap ({chunk_overlap}) cannot be >= chunk size ({chunk_size})"
            )

        self._chunk_size = chunk_size
        self._chunk_overlap = chunk_overlap
        self._keep_separator = keep_separator

        self._separators = separators or [
            "\n\n",
            "\n",
            " ",
            "",
        ]

    def split_text(self, text: str) -> list[str]:
        return self._split_text(text, self._separators)

    def _split_text(self, text: str, separators: list[str]) -> list[str]:
        separator = separators[-1]
        new_separators = []

        for i, sep in enumerate(separators):
            if sep == "":
                separator = sep
                break
            if re.search(re.escape(sep), text):
                separator = sep
                new_separators = separators[i + 1 :]
                break

        splits = self._split_text_with_separator(text, separator)

        good_splits = []

        for split in splits:
            if len(split) < self._chunk_size:
                good_splits.append(split)
            else:
                if new_separators:
                    other_info = self._split_text(split, new_separators)
                    good_splits.extend(other_info)
                else:
                    good_splits.extend(self._split_by_characters(split))

        return self._merge_splits(good_splits, separator)

    def _split_text_with_separator(self, text: str, separator: str) -> list[str]:
        if separator == "":
            return list(text)

        if self._keep_separator and separator in text:
            parts = text.split(separator)
            splits = []

            for i, part in enumerate(parts):
                if i == 0:
                    splits.append(part)
                elif i == len(parts) - 1:
                    if part:
                        splits.append(separator + part)
                else:
                    if part:
                        splits.append(separator + part)
                    else:
                        if splits:
                            splits[-1] += separator

            return [s for s in splits if s]
        return text.split(separator)

    def _split_by_characters(self, text: str) -> list[str]:
        chunks = []
        for i in range(0, len(text), self._chunk_size):
            chunks.append(text[i : i + self._chunk_size])  # noqa: PERF401
        return chunks

    def _merge_splits(self, splits: list[str], separator: str) -> list[str]:
        """Merge splits into chunks with proper overlap."""
        docs = []
        current_doc = []
        total = 0

        for split in splits:
            split_len = len(split)

            if total + split_len > self._chunk_size and current_doc:
                if separator == "":
                    doc = "".join(current_doc)
                else:
                    if self._keep_separator and separator == " ":
                        doc = "".join(current_doc)
                    else:
                        doc = separator.join(current_doc)

                if doc:
                    docs.append(doc)

                # Handle overlap by keeping some of the previous content
                while total > self._chunk_overlap and len(current_doc) > 1:
                    removed = current_doc.pop(0)
                    total -= len(removed)
                    if separator != "":
                        total -= len(separator)

            current_doc.append(split)
            total += split_len
            if separator != "" and len(current_doc) > 1:
                total += len(separator)

        if current_doc:
            if separator == "":
                doc = "".join(current_doc)
            else:
                if self._keep_separator and separator == " ":
                    doc = "".join(current_doc)
                else:
                    doc = separator.join(current_doc)

            if doc:
                docs.append(doc)

        return docs


class BaseChunker:
    def __init__(
        self,
        chunk_size: int = 1000,
        chunk_overlap: int = 200,
        separators: list[str] | None = None,
        keep_separator: bool = True,
    ):
        """Initialize the Chunker.

        Args:
            chunk_size: Maximum size of each chunk
            chunk_overlap: Number of characters to overlap between chunks
            separators: List of separators to use for splitting
            keep_separator: Whether to keep separators in the chunks
        """
        self._splitter = RecursiveCharacterTextSplitter(
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap,
            separators=separators,
            keep_separator=keep_separator,
        )

    def chunk(self, text: str) -> list[str]:
        if not text or not text.strip():
            return []

        return self._splitter.split_text(text)



import random

from crewai import Agent, Crew, Task
from patronus import Client, EvaluationResult
from patronus_local_evaluator_tool import PatronusLocalEvaluatorTool


# Test the PatronusLocalEvaluatorTool where agent uses the local evaluator
client = Client()


# Example of an evaluator that returns a random pass/fail result
@client.register_local_evaluator("random_evaluator")
def random_evaluator(**kwargs):
    score = random.random()
    return EvaluationResult(
        score_raw=score,
        pass_=score >= 0.5,
        explanation="example explanation",  # Optional justification for LLM judges
    )


# 1. Uses PatronusEvalTool: agent can pick the best evaluator and criteria
# patronus_eval_tool = PatronusEvalTool()

# 2. Uses PatronusPredefinedCriteriaEvalTool: agent uses the defined evaluator and criteria
# patronus_eval_tool = PatronusPredefinedCriteriaEvalTool(
#     evaluators=[{"evaluator": "judge", "criteria": "contains-code"}]
# )

# 3. Uses PatronusLocalEvaluatorTool: agent uses user defined evaluator
patronus_eval_tool = PatronusLocalEvaluatorTool(
    patronus_client=client,
    evaluator="random_evaluator",
    evaluated_model_gold_answer="example label",
)

# Create a new agent
coding_agent = Agent(
    role="Coding Agent",
    goal="Generate high quality code and verify that the output is code by using Patronus AI's evaluation tool.",
    backstory="You are an experienced coder who can generate high quality python code. You can follow complex instructions accurately and effectively.",
    tools=[patronus_eval_tool],
    verbose=True,
)

# Define tasks
generate_code = Task(
    description="Create a simple program to generate the first N numbers in the Fibonacci sequence. Select the most appropriate evaluator and criteria for evaluating your output.",
    expected_output="Program that generates the first N numbers in the Fibonacci sequence.",
    agent=coding_agent,
)

crew = Crew(agents=[coding_agent], tasks=[generate_code])

crew.kickoff()



import os
from typing import Any

from crewai.tools import BaseTool
from pydantic import BaseModel


def strtobool(val) -> bool:
    if isinstance(val, bool):
        return val
    val = val.lower()
    if val in ("y", "yes", "t", "true", "on", "1"):
        return True
    if val in ("n", "no", "f", "false", "off", "0"):
        return False
    raise ValueError(f"invalid value to cast to bool: {val!r}")


class FileWriterToolInput(BaseModel):
    filename: str
    directory: str | None = "./"
    overwrite: str | bool = False
    content: str


class FileWriterTool(BaseTool):
    name: str = "File Writer Tool"
    description: str = "A tool to write content to a specified file. Accepts filename, content, and optionally a directory path and overwrite flag as input."
    args_schema: type[BaseModel] = FileWriterToolInput

    def _run(self, **kwargs: Any) -> str:
        try:
            # Create the directory if it doesn't exist
            if kwargs.get("directory") and not os.path.exists(kwargs["directory"]):
                os.makedirs(kwargs["directory"])

            # Construct the full path
            filepath = os.path.join(kwargs.get("directory") or "", kwargs["filename"])

            # Convert overwrite to boolean
            kwargs["overwrite"] = strtobool(kwargs["overwrite"])

            # Check if file exists and overwrite is not allowed
            if os.path.exists(filepath) and not kwargs["overwrite"]:
                return f"File {filepath} already exists and overwrite option was not passed."

            # Write content to the file
            mode = "w" if kwargs["overwrite"] else "x"
            with open(filepath, mode) as file:
                file.write(kwargs["content"])
            return f"Content successfully written to {filepath}"
        except FileExistsError:
            return (
                f"File {filepath} already exists and overwrite option was not passed."
            )
        except KeyError as e:
            return f"An error occurred while accessing key: {e!s}"
        except Exception as e:
            return f"An error occurred while writing to the file: {e!s}"



import os
import unittest
from unittest.mock import MagicMock, patch

from crewai.tools import BaseTool
from crewai_tools.adapters.enterprise_adapter import EnterpriseActionTool
from crewai_tools.adapters.tool_collection import ToolCollection
from crewai_tools.tools import CrewaiEnterpriseTools


class TestCrewaiEnterpriseTools(unittest.TestCase):
    def setUp(self):
        self.mock_tools = [
            self._create_mock_tool("tool1", "Tool 1 Description"),
            self._create_mock_tool("tool2", "Tool 2 Description"),
            self._create_mock_tool("tool3", "Tool 3 Description"),
        ]
        self.adapter_patcher = patch(
            "crewai_tools.tools.crewai_enterprise_tools.crewai_enterprise_tools.EnterpriseActionKitToolAdapter"
        )
        self.MockAdapter = self.adapter_patcher.start()

        mock_adapter_instance = self.MockAdapter.return_value
        mock_adapter_instance.tools.return_value = self.mock_tools

    def tearDown(self):
        self.adapter_patcher.stop()

    def _create_mock_tool(self, name, description):
        mock_tool = MagicMock(spec=BaseTool)
        mock_tool.name = name
        mock_tool.description = description
        return mock_tool

    @patch.dict(os.environ, {"CREWAI_ENTERPRISE_TOOLS_TOKEN": "env-token"})
    def test_returns_tool_collection(self):
        tools = CrewaiEnterpriseTools()
        self.assertIsInstance(tools, ToolCollection)

    @patch.dict(os.environ, {"CREWAI_ENTERPRISE_TOOLS_TOKEN": "env-token"})
    def test_returns_all_tools_when_no_actions_list(self):
        tools = CrewaiEnterpriseTools()
        self.assertEqual(len(tools), 3)
        self.assertEqual(tools[0].name, "tool1")
        self.assertEqual(tools[1].name, "tool2")
        self.assertEqual(tools[2].name, "tool3")

    @patch.dict(os.environ, {"CREWAI_ENTERPRISE_TOOLS_TOKEN": "env-token"})
    def test_filters_tools_by_actions_list(self):
        tools = CrewaiEnterpriseTools(actions_list=["ToOl1", "tool3"])
        self.assertEqual(len(tools), 2)
        self.assertEqual(tools[0].name, "tool1")
        self.assertEqual(tools[1].name, "tool3")

    def test_uses_provided_parameters(self):
        CrewaiEnterpriseTools(
            enterprise_token="test-token",
            enterprise_action_kit_project_id="project-id",
            enterprise_action_kit_project_url="project-url",
        )

        self.MockAdapter.assert_called_once_with(
            enterprise_action_token="test-token",
            enterprise_action_kit_project_id="project-id",
            enterprise_action_kit_project_url="project-url",
        )

    @patch.dict(os.environ, {"CREWAI_ENTERPRISE_TOOLS_TOKEN": "env-token"})
    def test_uses_environment_token(self):
        CrewaiEnterpriseTools()
        self.MockAdapter.assert_called_once_with(enterprise_action_token="env-token")

    @patch.dict(os.environ, {"CREWAI_ENTERPRISE_TOOLS_TOKEN": "env-token"})
    def test_uses_environment_token_when_no_token_provided(self):
        CrewaiEnterpriseTools(enterprise_token="")
        self.MockAdapter.assert_called_once_with(enterprise_action_token="env-token")

    @patch.dict(
        os.environ,
        {
            "CREWAI_ENTERPRISE_TOOLS_TOKEN": "env-token",
            "CREWAI_ENTERPRISE_TOOLS_ACTIONS_LIST": '["tool1", "tool3"]',
        },
    )
    def test_uses_environment_actions_list(self):
        tools = CrewaiEnterpriseTools()
        self.assertEqual(len(tools), 2)
        self.assertEqual(tools[0].name, "tool1")
        self.assertEqual(tools[1].name, "tool3")


class TestEnterpriseActionToolSchemaConversion(unittest.TestCase):
    """Test the enterprise action tool schema conversion and validation."""

    def setUp(self):
        self.test_schema = {
            "type": "function",
            "function": {
                "name": "TEST_COMPLEX_ACTION",
                "description": "Test action with complex nested structure",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "filterCriteria": {
                            "type": "object",
                            "description": "Filter criteria object",
                            "properties": {
                                "operation": {"type": "string", "enum": ["AND", "OR"]},
                                "rules": {
                                    "type": "array",
                                    "items": {
                                        "type": "object",
                                        "properties": {
                                            "field": {
                                                "type": "string",
                                                "enum": ["name", "email", "status"],
                                            },
                                            "operator": {
                                                "type": "string",
                                                "enum": ["equals", "contains"],
                                            },
                                            "value": {"type": "string"},
                                        },
                                        "required": ["field", "operator", "value"],
                                    },
                                },
                            },
                            "required": ["operation", "rules"],
                        },
                        "options": {
                            "type": "object",
                            "properties": {
                                "limit": {"type": "integer"},
                                "offset": {"type": "integer"},
                            },
                            "required": [],
                        },
                    },
                    "required": [],
                },
            },
        }

    def test_complex_schema_conversion(self):
        """Test that complex nested schemas are properly converted to Pydantic models."""
        tool = EnterpriseActionTool(
            name="gmail_search_for_email",
            description="Test tool",
            enterprise_action_token="test_token",
            action_name="GMAIL_SEARCH_FOR_EMAIL",
            action_schema=self.test_schema,
        )

        self.assertEqual(tool.name, "gmail_search_for_email")
        self.assertEqual(tool.action_name, "GMAIL_SEARCH_FOR_EMAIL")

        schema_class = tool.args_schema
        self.assertIsNotNone(schema_class)

        schema_fields = schema_class.model_fields
        self.assertIn("filterCriteria", schema_fields)
        self.assertIn("options", schema_fields)

        # Test valid input structure
        valid_input = {
            "filterCriteria": {
                "operation": "AND",
                "rules": [
                    {"field": "name", "operator": "contains", "value": "test"},
                    {"field": "status", "operator": "equals", "value": "active"},
                ],
            },
            "options": {"limit": 10},
        }

        # This should not raise an exception
        validated_input = schema_class(**valid_input)
        self.assertIsNotNone(validated_input.filterCriteria)
        self.assertIsNotNone(validated_input.options)

    def test_optional_fields_validation(self):
        """Test that optional fields work correctly."""
        tool = EnterpriseActionTool(
            name="gmail_search_for_email",
            description="Test tool",
            enterprise_action_token="test_token",
            action_name="GMAIL_SEARCH_FOR_EMAIL",
            action_schema=self.test_schema,
        )

        schema_class = tool.args_schema

        minimal_input = {}
        validated_input = schema_class(**minimal_input)
        self.assertIsNone(validated_input.filterCriteria)
        self.assertIsNone(validated_input.options)

        partial_input = {"options": {"limit": 10}}
        validated_input = schema_class(**partial_input)
        self.assertIsNone(validated_input.filterCriteria)
        self.assertIsNotNone(validated_input.options)

    def test_enum_validation(self):
        """Test that enum values are properly validated."""
        tool = EnterpriseActionTool(
            name="gmail_search_for_email",
            description="Test tool",
            enterprise_action_token="test_token",
            action_name="GMAIL_SEARCH_FOR_EMAIL",
            action_schema=self.test_schema,
        )

        schema_class = tool.args_schema

        invalid_input = {
            "filterCriteria": {
                "operation": "INVALID_OPERATOR",
                "rules": [],
            }
        }

        with self.assertRaises(Exception):
            schema_class(**invalid_input)

    def test_required_nested_fields(self):
        """Test that required fields in nested objects are validated."""
        tool = EnterpriseActionTool(
            name="gmail_search_for_email",
            description="Test tool",
            enterprise_action_token="test_token",
            action_name="GMAIL_SEARCH_FOR_EMAIL",
            action_schema=self.test_schema,
        )

        schema_class = tool.args_schema

        incomplete_input = {
            "filterCriteria": {
                "operation": "OR",
                "rules": [
                    {
                        "field": "name",
                        "operator": "contains",
                    }
                ],
            }
        }

        with self.assertRaises(Exception):
            schema_class(**incomplete_input)

    @patch("requests.post")
    def test_tool_execution_with_complex_input(self, mock_post):
        """Test that the tool can execute with complex validated input."""
        mock_response = MagicMock()
        mock_response.ok = True
        mock_response.json.return_value = {"success": True, "results": []}
        mock_post.return_value = mock_response

        tool = EnterpriseActionTool(
            name="gmail_search_for_email",
            description="Test tool",
            enterprise_action_token="test_token",
            action_name="GMAIL_SEARCH_FOR_EMAIL",
            action_schema=self.test_schema,
        )

        tool._run(
            filterCriteria={
                "operation": "OR",
                "rules": [
                    {"field": "name", "operator": "contains", "value": "test"},
                    {"field": "status", "operator": "equals", "value": "active"},
                ],
            },
            options={"limit": 10},
        )

        mock_post.assert_called_once()
        call_args = mock_post.call_args
        payload = call_args[1]["json"]

        self.assertIn("filterCriteria", payload)
        self.assertIn("options", payload)
        self.assertEqual(payload["filterCriteria"]["operation"], "OR")

    def test_model_naming_convention(self):
        """Test that generated model names follow proper conventions."""
        tool = EnterpriseActionTool(
            name="gmail_search_for_email",
            description="Test tool",
            enterprise_action_token="test_token",
            action_name="GMAIL_SEARCH_FOR_EMAIL",
            action_schema=self.test_schema,
        )

        schema_class = tool.args_schema
        self.assertIsNotNone(schema_class)

        self.assertTrue(schema_class.__name__.endswith("Schema"))
        self.assertTrue(schema_class.__name__[0].isupper())

        complex_input = {
            "filterCriteria": {
                "operation": "OR",
                "rules": [
                    {"field": "name", "operator": "contains", "value": "test"},
                    {"field": "status", "operator": "equals", "value": "active"},
                ],
            },
            "options": {"limit": 10},
        }

        validated = schema_class(**complex_input)
        self.assertIsNotNone(validated.filterCriteria)

    def test_simple_schema_with_enums(self):
        """Test a simpler schema with basic enum validation."""
        simple_schema = {
            "type": "function",
            "function": {
                "name": "SIMPLE_TEST",
                "description": "Simple test function",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "status": {
                            "type": "string",
                            "enum": ["active", "inactive", "pending"],
                        },
                        "priority": {"type": "integer", "enum": [1, 2, 3]},
                    },
                    "required": ["status"],
                },
            },
        }

        tool = EnterpriseActionTool(
            name="simple_test",
            description="Simple test tool",
            enterprise_action_token="test_token",
            action_name="SIMPLE_TEST",
            action_schema=simple_schema,
        )

        schema_class = tool.args_schema

        valid_input = {"status": "active", "priority": 2}
        validated = schema_class(**valid_input)
        self.assertEqual(validated.status, "active")
        self.assertEqual(validated.priority, 2)

        with self.assertRaises(Exception):
            schema_class(status="invalid_status")



import os
import tempfile
from unittest.mock import Mock, patch

from crewai_tools.rag.base_loader import LoaderResult
from crewai_tools.rag.loaders.mdx_loader import MDXLoader
from crewai_tools.rag.source_content import SourceContent
import pytest


class TestMDXLoader:
    def _write_temp_mdx(self, content):
        f = tempfile.NamedTemporaryFile(mode="w", suffix=".mdx", delete=False)
        f.write(content)
        f.close()
        return f.name

    def _load_from_file(self, content):
        path = self._write_temp_mdx(content)
        try:
            loader = MDXLoader()
            return loader.load(SourceContent(path)), path
        finally:
            os.unlink(path)

    def test_load_basic_mdx_file(self):
        content = """
import Component from './Component'
export const meta = { title: 'Test' }

# Test MDX File

This is a **markdown** file with JSX.

<Component prop="value" />

Some more content.

<div className="container">
    <p>Nested content</p>
</div>
"""
        result, path = self._load_from_file(content)

        assert isinstance(result, LoaderResult)
        assert all(
            tag not in result.content
            for tag in ["import", "export", "<Component", "<div", "</div>"]
        )
        assert all(
            text in result.content
            for text in [
                "# Test MDX File",
                "markdown",
                "Some more content",
                "Nested content",
            ]
        )
        assert result.metadata["format"] == "mdx"
        assert result.source == path

    def test_mdx_multiple_imports_exports(self):
        content = """
import React from 'react'
import { useState } from 'react'
import CustomComponent from './custom'

export default function Layout() { return null }
export const config = { test: true }

# Content

Regular markdown content here.
"""
        result, _ = self._load_from_file(content)
        assert "# Content" in result.content
        assert "Regular markdown content here." in result.content
        assert "import" not in result.content and "export" not in result.content

    def test_complex_jsx_cleanup(self):
        content = """
# MDX with Complex JSX

<div className="alert alert-info">
    <strong>Info:</strong> This is important information.
    <ul><li>Item 1</li><li>Item 2</li></ul>
</div>

Regular paragraph text.

<MyComponent prop1="value1">Nested content inside component</MyComponent>
"""
        result, _ = self._load_from_file(content)
        assert all(
            tag not in result.content
            for tag in ["<div", "<strong>", "<ul>", "<MyComponent"]
        )
        assert all(
            text in result.content
            for text in [
                "Info:",
                "Item 1",
                "Regular paragraph text.",
                "Nested content inside component",
            ]
        )

    def test_whitespace_cleanup(self):
        content = """


# Title


Some content.


More content after multiple newlines.



Final content.
"""
        result, _ = self._load_from_file(content)
        assert result.content.count("\n\n\n") == 0
        assert result.content.startswith("# Title")
        assert result.content.endswith("Final content.")

    def test_only_jsx_content(self):
        content = """
<div>
    <h1>Only JSX content</h1>
    <p>No markdown here</p>
</div>
"""
        result, _ = self._load_from_file(content)
        assert all(tag not in result.content for tag in ["<div>", "<h1>", "<p>"])
        assert "Only JSX content" in result.content
        assert "No markdown here" in result.content

    @patch("requests.get")
    def test_load_mdx_from_url(self, mock_get):
        mock_get.return_value = Mock(
            text="# MDX from URL\n\nContent here.\n\n<Component />",
            raise_for_status=lambda: None,
        )
        loader = MDXLoader()
        result = loader.load(SourceContent("https://example.com/content.mdx"))
        assert "# MDX from URL" in result.content
        assert "<Component />" not in result.content

    @patch("requests.get")
    def test_load_mdx_with_custom_headers(self, mock_get):
        mock_get.return_value = Mock(
            text="# Custom headers test", raise_for_status=lambda: None
        )
        loader = MDXLoader()
        loader.load(
            SourceContent("https://example.com"),
            headers={"Authorization": "Bearer token"},
        )
        assert mock_get.call_args[1]["headers"] == {"Authorization": "Bearer token"}

    @patch("requests.get")
    def test_mdx_url_fetch_error(self, mock_get):
        mock_get.side_effect = Exception("Network error")
        with pytest.raises(ValueError, match="Error fetching MDX from URL"):
            MDXLoader().load(SourceContent("https://example.com"))

    def test_load_inline_mdx_text(self):
        content = """# Inline MDX\n\nimport Something from 'somewhere'\n\nContent with <Component prop=\"value\" />.\n\nexport const meta = { title: 'Test' }"""
        loader = MDXLoader()
        result = loader.load(SourceContent(content))
        assert "# Inline MDX" in result.content
        assert "Content with ." in result.content

    def test_empty_result_after_cleaning(self):
        content = """
import Something from 'somewhere'
export const config = {}
<div></div>
"""
        result, _ = self._load_from_file(content)
        assert result.content.strip() == ""

    def test_edge_case_parsing(self):
        content = """
# Title

<Component>
Multi-line
JSX content
</Component>

import { a, b } from 'module'

export { x, y }

Final text.
"""
        result, _ = self._load_from_file(content)
        assert "# Title" in result.content
        assert "JSX content" in result.content
        assert "Final text." in result.content
        assert all(
            phrase not in result.content
            for phrase in ["import {", "export {", "<Component>"]
        )



from pydantic import BaseModel, Field

from crewai_tools.rag.data_types import DataType

from ..rag.rag_tool import RagTool


class FixedGithubSearchToolSchema(BaseModel):
    """Input for GithubSearchTool."""

    search_query: str = Field(
        ...,
        description="Mandatory search query you want to use to search the github repo's content",
    )


class GithubSearchToolSchema(FixedGithubSearchToolSchema):
    """Input for GithubSearchTool."""

    github_repo: str = Field(..., description="Mandatory github you want to search")
    content_types: list[str] = Field(
        ...,
        description="Mandatory content types you want to be included search, options: [code, repo, pr, issue]",
    )


class GithubSearchTool(RagTool):
    name: str = "Search a github repo's content"
    description: str = "A tool that can be used to semantic search a query from a github repo's content. This is not the GitHub API, but instead a tool that can provide semantic search capabilities."
    summarize: bool = False
    gh_token: str
    args_schema: type[BaseModel] = GithubSearchToolSchema
    content_types: list[str] = Field(
        default_factory=lambda: ["code", "repo", "pr", "issue"],
        description="Content types you want to be included search, options: [code, repo, pr, issue]",
    )

    def __init__(
        self,
        github_repo: str | None = None,
        content_types: list[str] | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        if github_repo and content_types:
            self.add(repo=github_repo, content_types=content_types)
            self.description = f"A tool that can be used to semantic search a query the {github_repo} github repo's content. This is not the GitHub API, but instead a tool that can provide semantic search capabilities."
            self.args_schema = FixedGithubSearchToolSchema
            self._generate_description()

    def add(
        self,
        repo: str,
        content_types: list[str] | None = None,
    ) -> None:
        content_types = content_types or self.content_types
        super().add(
            f"https://github.com/{repo}",
            data_type=DataType.GITHUB,
            metadata={"content_types": content_types, "gh_token": self.gh_token},
        )

    def _run(
        self,
        search_query: str,
        github_repo: str | None = None,
        content_types: list[str] | None = None,
        similarity_threshold: float | None = None,
        limit: int | None = None,
    ) -> str:
        if github_repo:
            self.add(
                repo=github_repo,
                content_types=content_types,
            )
        return super()._run(
            query=search_query, similarity_threshold=similarity_threshold, limit=limit
        )



"""PDF loader for extracting text from PDF files."""

import os
from pathlib import Path
from typing import Any

from crewai_tools.rag.base_loader import BaseLoader, LoaderResult
from crewai_tools.rag.source_content import SourceContent


class PDFLoader(BaseLoader):
    """Loader for PDF files."""

    def load(self, source: SourceContent, **kwargs) -> LoaderResult:
        """Load and extract text from a PDF file.

        Args:
            source: The source content containing the PDF file path

        Returns:
            LoaderResult with extracted text content

        Raises:
            FileNotFoundError: If the PDF file doesn't exist
            ImportError: If required PDF libraries aren't installed
        """
        try:
            import pypdf
        except ImportError:
            try:
                import PyPDF2 as pypdf  # noqa: N813
            except ImportError as e:
                raise ImportError(
                    "PDF support requires pypdf or PyPDF2. Install with: uv add pypdf"
                ) from e

        file_path = source.source

        if not os.path.isfile(file_path):
            raise FileNotFoundError(f"PDF file not found: {file_path}")

        text_content = []
        metadata: dict[str, Any] = {
            "source": str(file_path),
            "file_name": Path(file_path).name,
            "file_type": "pdf",
        }

        try:
            with open(file_path, "rb") as file:
                pdf_reader = pypdf.PdfReader(file)
                metadata["num_pages"] = len(pdf_reader.pages)

                for page_num, page in enumerate(pdf_reader.pages, 1):
                    page_text = page.extract_text()
                    if page_text.strip():
                        text_content.append(f"Page {page_num}:\n{page_text}")
        except Exception as e:
            raise ValueError(f"Error reading PDF file {file_path}: {e!s}") from e

        if not text_content:
            content = f"[PDF file with no extractable text: {Path(file_path).name}]"
        else:
            content = "\n\n".join(text_content)

        return LoaderResult(
            content=content,
            source=str(file_path),
            metadata=metadata,
            doc_id=self.generate_doc_id(source_ref=str(file_path), content=content),
        )



import os
from typing import Any

from crewai.tools import BaseTool, EnvVar
from pydantic import BaseModel, Field


class BrowserbaseLoadToolSchema(BaseModel):
    url: str = Field(description="Website URL")


class BrowserbaseLoadTool(BaseTool):
    name: str = "Browserbase web load tool"
    description: str = "Load webpages url in a headless browser using Browserbase and return the contents"
    args_schema: type[BaseModel] = BrowserbaseLoadToolSchema
    api_key: str | None = os.getenv("BROWSERBASE_API_KEY")
    project_id: str | None = os.getenv("BROWSERBASE_PROJECT_ID")
    text_content: bool | None = False
    session_id: str | None = None
    proxy: bool | None = None
    browserbase: Any | None = None
    package_dependencies: list[str] = Field(default_factory=lambda: ["browserbase"])
    env_vars: list[EnvVar] = Field(
        default_factory=lambda: [
            EnvVar(
                name="BROWSERBASE_API_KEY",
                description="API key for Browserbase services",
                required=False,
            ),
            EnvVar(
                name="BROWSERBASE_PROJECT_ID",
                description="Project ID for Browserbase services",
                required=False,
            ),
        ]
    )

    def __init__(
        self,
        api_key: str | None = None,
        project_id: str | None = None,
        text_content: bool | None = False,
        session_id: str | None = None,
        proxy: bool | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if not self.api_key:
            raise EnvironmentError(
                "BROWSERBASE_API_KEY environment variable is required for initialization"
            )
        try:
            from browserbase import Browserbase  # type: ignore
        except ImportError:
            import click

            if click.confirm(
                "`browserbase` package not found, would you like to install it?"
            ):
                import subprocess

                subprocess.run(["uv", "add", "browserbase"], check=True)  # noqa: S607
                from browserbase import Browserbase  # type: ignore
            else:
                raise ImportError(
                    "`browserbase` package not found, please run `uv add browserbase`"
                ) from None

        self.browserbase = Browserbase(api_key=self.api_key)
        self.text_content = text_content
        self.session_id = session_id
        self.proxy = proxy

    def _run(self, url: str):
        return self.browserbase.load_url(
            url, self.text_content, self.session_id, self.proxy
        )



from typing import Any

from crewai.tools import BaseTool
import requests

from crewai_tools.tools.crewai_platform_tools.crewai_platform_action_tool import (
    CrewAIPlatformActionTool,
)
from crewai_tools.tools.crewai_platform_tools.misc import (
    get_platform_api_base_url,
    get_platform_integration_token,
)


class CrewaiPlatformToolBuilder:
    def __init__(
        self,
        apps: list[str],
    ):
        self._apps = apps
        self._actions_schema = {}
        self._tools = None

    def tools(self) -> list[BaseTool]:
        if self._tools is None:
            self._fetch_actions()
            self._create_tools()
        return self._tools if self._tools is not None else []

    def _fetch_actions(self):
        actions_url = f"{get_platform_api_base_url()}/actions"
        headers = {"Authorization": f"Bearer {get_platform_integration_token()}"}

        try:
            response = requests.get(
                actions_url,
                headers=headers,
                timeout=30,
                params={"apps": ",".join(self._apps)},
            )
            response.raise_for_status()
        except Exception:
            return

        raw_data = response.json()

        self._actions_schema = {}
        action_categories = raw_data.get("actions", {})

        for app, action_list in action_categories.items():
            if isinstance(action_list, list):
                for action in action_list:
                    if action_name := action.get("name"):
                        action_schema = {
                            "function": {
                                "name": action_name,
                                "description": action.get(
                                    "description", f"Execute {action_name}"
                                ),
                                "parameters": action.get("parameters", {}),
                                "app": app,
                            }
                        }
                        self._actions_schema[action_name] = action_schema

    def _generate_detailed_description(
        self, schema: dict[str, Any], indent: int = 0
    ) -> list[str]:
        descriptions = []
        indent_str = "  " * indent

        schema_type = schema.get("type", "string")

        if schema_type == "object":
            properties = schema.get("properties", {})
            required_fields = schema.get("required", [])

            if properties:
                descriptions.append(f"{indent_str}Object with properties:")
                for prop_name, prop_schema in properties.items():
                    prop_desc = prop_schema.get("description", "")
                    is_required = prop_name in required_fields
                    req_str = " (required)" if is_required else " (optional)"
                    descriptions.append(
                        f"{indent_str}  - {prop_name}: {prop_desc}{req_str}"
                    )

                    if prop_schema.get("type") == "object":
                        descriptions.extend(
                            self._generate_detailed_description(prop_schema, indent + 2)
                        )
                    elif prop_schema.get("type") == "array":
                        items_schema = prop_schema.get("items", {})
                        if items_schema.get("type") == "object":
                            descriptions.append(f"{indent_str}    Array of objects:")
                            descriptions.extend(
                                self._generate_detailed_description(
                                    items_schema, indent + 3
                                )
                            )
                        elif "enum" in items_schema:
                            descriptions.append(
                                f"{indent_str}    Array of enum values: {items_schema['enum']}"
                            )
                    elif "enum" in prop_schema:
                        descriptions.append(
                            f"{indent_str}    Enum values: {prop_schema['enum']}"
                        )

        return descriptions

    def _create_tools(self):
        tools = []

        for action_name, action_schema in self._actions_schema.items():
            function_details = action_schema.get("function", {})
            description = function_details.get("description", f"Execute {action_name}")

            parameters = function_details.get("parameters", {})
            param_descriptions = []

            if parameters.get("properties"):
                param_descriptions.append("\nDetailed Parameter Structure:")
                param_descriptions.extend(
                    self._generate_detailed_description(parameters)
                )

            full_description = description + "\n".join(param_descriptions)

            tool = CrewAIPlatformActionTool(
                description=full_description,
                action_name=action_name,
                action_schema=action_schema,
            )

            tools.append(tool)

        self._tools = tools

    def __enter__(self):
        return self.tools()

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass



from pydantic import BaseModel, Field

from crewai_tools.rag.data_types import DataType

from ..rag.rag_tool import RagTool


class FixedMDXSearchToolSchema(BaseModel):
    """Input for MDXSearchTool."""

    search_query: str = Field(
        ...,
        description="Mandatory search query you want to use to search the MDX's content",
    )


class MDXSearchToolSchema(FixedMDXSearchToolSchema):
    """Input for MDXSearchTool."""

    mdx: str = Field(..., description="File path or URL of a MDX file to be searched")


class MDXSearchTool(RagTool):
    name: str = "Search a MDX's content"
    description: str = (
        "A tool that can be used to semantic search a query from a MDX's content."
    )
    args_schema: type[BaseModel] = MDXSearchToolSchema

    def __init__(self, mdx: str | None = None, **kwargs):
        super().__init__(**kwargs)
        if mdx is not None:
            self.add(mdx)
            self.description = f"A tool that can be used to semantic search a query the {mdx} MDX's content."
            self.args_schema = FixedMDXSearchToolSchema
            self._generate_description()

    def add(self, mdx: str) -> None:
        super().add(mdx, data_type=DataType.MDX)

    def _run(
        self,
        search_query: str,
        mdx: str | None = None,
        similarity_threshold: float | None = None,
        limit: int | None = None,
    ) -> str:
        if mdx is not None:
            self.add(mdx)
        return super()._run(
            query=search_query, similarity_threshold=similarity_threshold, limit=limit
        )



import logging
from pathlib import Path
import re
import time
from typing import ClassVar
import urllib.error
import urllib.parse
import urllib.request
import xml.etree.ElementTree as ET

from crewai.tools import BaseTool, EnvVar
from pydantic import BaseModel, ConfigDict, Field


logger = logging.getLogger(__file__)


class ArxivToolInput(BaseModel):
    search_query: str = Field(
        ..., description="Search query for Arxiv, e.g., 'transformer neural network'"
    )
    max_results: int = Field(
        5, ge=1, le=100, description="Max results to fetch; must be between 1 and 100"
    )


class ArxivPaperTool(BaseTool):
    BASE_API_URL: ClassVar[str] = "http://export.arxiv.org/api/query"
    SLEEP_DURATION: ClassVar[int] = 1
    SUMMARY_TRUNCATE_LENGTH: ClassVar[int] = 300
    ATOM_NAMESPACE: ClassVar[str] = "{http://www.w3.org/2005/Atom}"
    REQUEST_TIMEOUT: ClassVar[int] = 10
    name: str = "Arxiv Paper Fetcher and Downloader"
    description: str = "Fetches metadata from Arxiv based on a search query and optionally downloads PDFs."
    args_schema: type[BaseModel] = ArxivToolInput
    model_config = ConfigDict(extra="allow")
    package_dependencies: list[str] = Field(default_factory=lambda: ["pydantic"])
    env_vars: list[EnvVar] = Field(default_factory=list)

    def __init__(
        self, download_pdfs=False, save_dir="./arxiv_pdfs", use_title_as_filename=False
    ):
        super().__init__()
        self.download_pdfs = download_pdfs
        self.save_dir = save_dir
        self.use_title_as_filename = use_title_as_filename

    def _run(self, search_query: str, max_results: int = 5) -> str:
        try:
            args = ArxivToolInput(search_query=search_query, max_results=max_results)
            logger.info(
                f"Running Arxiv tool: query='{args.search_query}', max_results={args.max_results}, "
                f"download_pdfs={self.download_pdfs}, save_dir='{self.save_dir}', "
                f"use_title_as_filename={self.use_title_as_filename}"
            )

            papers = self.fetch_arxiv_data(args.search_query, args.max_results)

            if self.download_pdfs:
                save_dir = self._validate_save_path(self.save_dir)
                for paper in papers:
                    if paper["pdf_url"]:
                        if self.use_title_as_filename:
                            safe_title = re.sub(
                                r'[\\/*?:"<>|]', "_", paper["title"]
                            ).strip()
                            filename_base = safe_title or paper["arxiv_id"]
                        else:
                            filename_base = paper["arxiv_id"]
                        filename = f"{filename_base[:500]}.pdf"
                        save_path = Path(save_dir) / filename

                        self.download_pdf(paper["pdf_url"], save_path)
                        time.sleep(self.SLEEP_DURATION)

            results = [self._format_paper_result(p) for p in papers]
            return "\n\n" + "-" * 80 + "\n\n".join(results)

        except Exception as e:
            logger.error(f"ArxivTool Error: {e!s}")
            return f"Failed to fetch or download Arxiv papers: {e!s}"

    def fetch_arxiv_data(self, search_query: str, max_results: int) -> list[dict]:
        api_url = f"{self.BASE_API_URL}?search_query={urllib.parse.quote(search_query)}&start=0&max_results={max_results}"
        logger.info(f"Fetching data from Arxiv API: {api_url}")

        try:
            with urllib.request.urlopen(  # noqa: S310
                api_url, timeout=self.REQUEST_TIMEOUT
            ) as response:
                if response.status != 200:
                    raise Exception(f"HTTP {response.status}: {response.reason}")
                data = response.read().decode("utf-8")
        except urllib.error.URLError as e:
            logger.error(f"Error fetching data from Arxiv: {e}")
            raise

        root = ET.fromstring(data)  # noqa: S314
        papers = []

        for entry in root.findall(self.ATOM_NAMESPACE + "entry"):
            raw_id = self._get_element_text(entry, "id")
            arxiv_id = raw_id.split("/")[-1].replace(".", "_") if raw_id else "unknown"

            title = self._get_element_text(entry, "title") or "No Title"
            summary = self._get_element_text(entry, "summary") or "No Summary"
            published = self._get_element_text(entry, "published") or "No Publish Date"
            authors = [
                self._get_element_text(author, "name") or "Unknown"
                for author in entry.findall(self.ATOM_NAMESPACE + "author")
            ]

            pdf_url = self._extract_pdf_url(entry)

            papers.append(
                {
                    "arxiv_id": arxiv_id,
                    "title": title,
                    "summary": summary,
                    "authors": authors,
                    "published_date": published,
                    "pdf_url": pdf_url,
                }
            )

        return papers

    @staticmethod
    def _get_element_text(entry: ET.Element, element_name: str) -> str | None:
        elem = entry.find(f"{ArxivPaperTool.ATOM_NAMESPACE}{element_name}")
        return elem.text.strip() if elem is not None and elem.text else None

    def _extract_pdf_url(self, entry: ET.Element) -> str | None:
        for link in entry.findall(self.ATOM_NAMESPACE + "link"):
            if link.attrib.get("title", "").lower() == "pdf":
                return link.attrib.get("href")
        for link in entry.findall(self.ATOM_NAMESPACE + "link"):
            href = link.attrib.get("href")
            if href and "pdf" in href:
                return href
        return None

    def _format_paper_result(self, paper: dict) -> str:
        summary = (
            (paper["summary"][: self.SUMMARY_TRUNCATE_LENGTH] + "...")
            if len(paper["summary"]) > self.SUMMARY_TRUNCATE_LENGTH
            else paper["summary"]
        )
        authors_str = ", ".join(paper["authors"])
        return (
            f"Title: {paper['title']}\n"
            f"Authors: {authors_str}\n"
            f"Published: {paper['published_date']}\n"
            f"PDF: {paper['pdf_url'] or 'N/A'}\n"
            f"Summary: {summary}"
        )

    @staticmethod
    def _validate_save_path(path: str) -> Path:
        save_path = Path(path).resolve()
        save_path.mkdir(parents=True, exist_ok=True)
        return save_path

    def download_pdf(self, pdf_url: str, save_path: str):
        try:
            logger.info(f"Downloading PDF from {pdf_url} to {save_path}")
            urllib.request.urlretrieve(pdf_url, str(save_path))  # noqa: S310
            logger.info(f"PDF saved: {save_path}")
        except urllib.error.URLError as e:
            logger.error(f"Network error occurred while downloading {pdf_url}: {e}")
            raise
        except OSError as e:
            logger.error(f"File save error for {save_path}: {e}")
            raise



from importlib.metadata import version
import json
import os
from platform import architecture, python_version
from typing import Any

from crewai.tools import BaseTool, EnvVar
from pydantic import BaseModel, ConfigDict, Field


try:
    from oxylabs import RealtimeClient
    from oxylabs.sources.response import Response as OxylabsResponse

    OXYLABS_AVAILABLE = True
except ImportError:
    RealtimeClient = Any
    OxylabsResponse = Any

    OXYLABS_AVAILABLE = False


__all__ = ["OxylabsAmazonSearchScraperConfig", "OxylabsAmazonSearchScraperTool"]


class OxylabsAmazonSearchScraperArgs(BaseModel):
    query: str = Field(description="Amazon search term")


class OxylabsAmazonSearchScraperConfig(BaseModel):
    """Amazon Search Scraper configuration options:
    https://developers.oxylabs.io/scraper-apis/web-scraper-api/targets/amazon/search.
    """

    domain: str | None = Field(
        None, description="The domain to limit the search results to."
    )
    start_page: int | None = Field(None, description="The starting page number.")
    pages: int | None = Field(None, description="The number of pages to scrape.")
    geo_location: str | None = Field(None, description="The Deliver to location.")
    user_agent_type: str | None = Field(None, description="Device type and browser.")
    render: str | None = Field(None, description="Enables JavaScript rendering.")
    callback_url: str | None = Field(None, description="URL to your callback endpoint.")
    context: list | None = Field(
        None,
        description="Additional advanced settings and controls for specialized requirements.",
    )
    parse: bool | None = Field(None, description="True will return structured data.")
    parsing_instructions: dict | None = Field(
        None, description="Instructions for parsing the results."
    )


class OxylabsAmazonSearchScraperTool(BaseTool):
    """Scrape Amazon search results with OxylabsAmazonSearchScraperTool.

    Get Oxylabs account:
    https://dashboard.oxylabs.io/en

    Args:
        username (str): Oxylabs username.
        password (str): Oxylabs password.
        config: Configuration options. See ``OxylabsAmazonSearchScraperConfig``
    """

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        validate_assignment=True,
    )
    name: str = "Oxylabs Amazon Search Scraper tool"
    description: str = "Scrape Amazon search results with Oxylabs Amazon Search Scraper"
    args_schema: type[BaseModel] = OxylabsAmazonSearchScraperArgs

    oxylabs_api: RealtimeClient
    config: OxylabsAmazonSearchScraperConfig
    package_dependencies: list[str] = Field(default_factory=lambda: ["oxylabs"])
    env_vars: list[EnvVar] = Field(
        default_factory=lambda: [
            EnvVar(
                name="OXYLABS_USERNAME",
                description="Username for Oxylabs",
                required=True,
            ),
            EnvVar(
                name="OXYLABS_PASSWORD",
                description="Password for Oxylabs",
                required=True,
            ),
        ]
    )

    def __init__(
        self,
        username: str | None = None,
        password: str | None = None,
        config: OxylabsAmazonSearchScraperConfig | dict | None = None,
        **kwargs,
    ):
        bits, _ = architecture()
        sdk_type = (
            f"oxylabs-crewai-sdk-python/"
            f"{version('crewai')} "
            f"({python_version()}; {bits})"
        )

        if username is None or password is None:
            username, password = self._get_credentials_from_env()

        if OXYLABS_AVAILABLE:
            # import RealtimeClient to make it accessible for the current scope
            from oxylabs import RealtimeClient

            kwargs["oxylabs_api"] = RealtimeClient(
                username=username,
                password=password,
                sdk_type=sdk_type,
            )
        else:
            import click

            if click.confirm(
                "You are missing the 'oxylabs' package. Would you like to install it?"
            ):
                import subprocess

                try:
                    subprocess.run(["uv", "add", "oxylabs"], check=True)  # noqa: S607
                    from oxylabs import RealtimeClient

                    kwargs["oxylabs_api"] = RealtimeClient(
                        username=username,
                        password=password,
                        sdk_type=sdk_type,
                    )
                except subprocess.CalledProcessError as e:
                    raise ImportError("Failed to install oxylabs package") from e
            else:
                raise ImportError(
                    "`oxylabs` package not found, please run `uv add oxylabs`"
                )

        if config is None:
            config = OxylabsAmazonSearchScraperConfig()
        super().__init__(config=config, **kwargs)

    def _get_credentials_from_env(self) -> tuple[str, str]:
        username = os.environ.get("OXYLABS_USERNAME")
        password = os.environ.get("OXYLABS_PASSWORD")
        if not username or not password:
            raise ValueError(
                "You must pass oxylabs username and password when instantiating the tool "
                "or specify OXYLABS_USERNAME and OXYLABS_PASSWORD environment variables"
            )
        return username, password

    def _run(self, query: str) -> str:
        response = self.oxylabs_api.amazon.scrape_search(
            query,
            **self.config.model_dump(exclude_none=True),
        )

        content = response.results[0].content

        if isinstance(content, dict):
            return json.dumps(content)

        return content



from collections.abc import Iterable
from importlib.metadata import version
from logging import getLogger
import os
from typing import Any

from crewai.tools import BaseTool, EnvVar
from openai import AzureOpenAI, Client
from pydantic import BaseModel, Field

from crewai_tools.tools.mongodb_vector_search_tool.utils import (
    create_vector_search_index,
)


try:
    import pymongo  # noqa: F401

    MONGODB_AVAILABLE = True
except ImportError:
    MONGODB_AVAILABLE = False

logger = getLogger(__name__)


class MongoDBVectorSearchConfig(BaseModel):
    """Configuration for MongoDB vector search queries."""

    limit: int | None = Field(default=4, description="number of documents to return.")
    pre_filter: dict[str, Any] | None = Field(
        default=None,
        description="List of MQL match expressions comparing an indexed field",
    )
    post_filter_pipeline: list[dict] | None = Field(
        default=None,
        description="Pipeline of MongoDB aggregation stages to filter/process results after $vectorSearch.",
    )
    oversampling_factor: int = Field(
        default=10,
        description="Multiple of limit used when generating number of candidates at each step in the HNSW Vector Search",
    )
    include_embeddings: bool = Field(
        default=False,
        description="Whether to include the embedding vector of each result in metadata.",
    )


class MongoDBToolSchema(BaseModel):
    """Input for MongoDBTool."""

    query: str = Field(
        ...,
        description="The query to search retrieve relevant information from the MongoDB database. Pass only the query, not the question.",
    )


class MongoDBVectorSearchTool(BaseTool):
    """Tool to perfrom a vector search the MongoDB database."""

    name: str = "MongoDBVectorSearchTool"
    description: str = "A tool to perfrom a vector search on a MongoDB database for relevant information on internal documents."

    args_schema: type[BaseModel] = MongoDBToolSchema
    query_config: MongoDBVectorSearchConfig | None = Field(
        default=None, description="MongoDB Vector Search query configuration"
    )
    embedding_model: str = Field(
        default="text-embedding-3-large",
        description="Text OpenAI embedding model to use",
    )
    vector_index_name: str = Field(
        default="vector_index", description="Name of the Atlas Search vector index"
    )
    text_key: str = Field(
        default="text",
        description="MongoDB field that will contain the text for each document",
    )
    embedding_key: str = Field(
        default="embedding",
        description="Field that will contain the embedding for each document",
    )
    database_name: str = Field(..., description="The name of the MongoDB database")
    collection_name: str = Field(..., description="The name of the MongoDB collection")
    connection_string: str = Field(
        ...,
        description="The connection string of the MongoDB cluster",
    )
    dimensions: int = Field(
        default=1536,
        description="Number of dimensions in the embedding vector",
    )
    env_vars: list[EnvVar] = Field(
        default_factory=lambda: [
            EnvVar(
                name="BROWSERBASE_API_KEY",
                description="API key for Browserbase services",
                required=False,
            ),
            EnvVar(
                name="BROWSERBASE_PROJECT_ID",
                description="Project ID for Browserbase services",
                required=False,
            ),
        ]
    )
    package_dependencies: list[str] = Field(default_factory=lambda: ["mongdb"])

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if not MONGODB_AVAILABLE:
            import click

            if click.confirm(
                "You are missing the 'mongodb' crewai tool. Would you like to install it?"
            ):
                import subprocess

                subprocess.run(["uv", "add", "pymongo"], check=True)  # noqa: S607

            else:
                raise ImportError("You are missing the 'mongodb' crewai tool.")

        if "AZURE_OPENAI_ENDPOINT" in os.environ:
            self._openai_client = AzureOpenAI()
        elif "OPENAI_API_KEY" in os.environ:
            self._openai_client = Client()
        else:
            raise ValueError(
                "OPENAI_API_KEY environment variable is required for MongoDBVectorSearchTool and it is mandatory to use the tool."
            )

        from pymongo import MongoClient
        from pymongo.driver_info import DriverInfo

        self._client = MongoClient(
            self.connection_string,
            driver=DriverInfo(name="CrewAI", version=version("crewai-tools")),
        )
        self._coll = self._client[self.database_name][self.collection_name]

    def create_vector_search_index(
        self,
        *,
        dimensions: int,
        relevance_score_fn: str = "cosine",
        auto_index_timeout: int = 15,
    ) -> None:
        """Convenience function to create a vector search index.

        Args:
            dimensions: Number of dimensions in embedding.  If the value is set and
                the index does not exist, an index will be created.
            relevance_score_fn: The similarity score used for the index
                Currently supported: 'euclidean', 'cosine', and 'dotProduct'
            auto_index_timeout: Timeout in seconds to wait for an auto-created index
               to be ready.
        """
        create_vector_search_index(
            collection=self._coll,
            index_name=self.vector_index_name,
            dimensions=dimensions,
            path=self.embedding_key,
            similarity=relevance_score_fn,
            wait_until_complete=auto_index_timeout,
        )

    def add_texts(
        self,
        texts: Iterable[str],
        metadatas: list[dict[str, Any]] | None = None,
        ids: list[str] | None = None,
        batch_size: int = 100,
        **kwargs: Any,
    ) -> list[str]:
        """Add texts, create embeddings, and add to the Collection and index.

        Important notes on ids:
            - If _id or id is a key in the metadatas dicts, one must
                pop them and provide as separate list.
            - They must be unique.
            - If they are not provided, the VectorStore will create unique ones,
                stored as bson.ObjectIds internally, and strings in Langchain.
                These will appear in Document.metadata with key, '_id'.

        Args:
            texts: Iterable of strings to add to the vectorstore.
            metadatas: Optional list of metadatas associated with the texts.
            ids: Optional list of unique ids that will be used as index in VectorStore.
                See note on ids.
            batch_size: Number of documents to insert at a time.
                Tuning this may help with performance and sidestep MongoDB limits.

        Returns:
            List of ids added to the vectorstore.
        """
        from bson import ObjectId

        _metadatas = metadatas or [{} for _ in texts]
        ids = [str(ObjectId()) for _ in range(len(list(texts)))]
        metadatas_batch = _metadatas

        result_ids = []
        texts_batch = []
        metadatas_batch = []
        size = 0
        i = 0
        for j, (text, metadata) in enumerate(zip(texts, _metadatas, strict=False)):
            size += len(text) + len(metadata)
            texts_batch.append(text)
            metadatas_batch.append(metadata)
            if (j + 1) % batch_size == 0 or size >= 47_000_000:
                batch_res = self._bulk_embed_and_insert_texts(
                    texts_batch, metadatas_batch, ids[i : j + 1]
                )
                result_ids.extend(batch_res)
                texts_batch = []
                metadatas_batch = []
                size = 0
                i = j + 1
        if texts_batch:
            batch_res = self._bulk_embed_and_insert_texts(
                texts_batch, metadatas_batch, ids[i : j + 1]
            )
            result_ids.extend(batch_res)
        return result_ids

    def _embed_texts(self, texts: list[str]) -> list[list[float]]:
        return [
            i.embedding
            for i in self._openai_client.embeddings.create(
                input=texts,
                model=self.embedding_model,
                dimensions=self.dimensions,
            ).data
        ]

    def _bulk_embed_and_insert_texts(
        self,
        texts: list[str],
        metadatas: list[dict],
        ids: list[str],
    ) -> list[str]:
        """Bulk insert single batch of texts, embeddings, and ids."""
        from bson import ObjectId
        from pymongo.operations import ReplaceOne

        if not texts:
            return []
        # Compute embedding vectors
        embeddings = self._embed_texts(texts)
        docs = [
            {
                "_id": ObjectId(i),
                self.text_key: t,
                self.embedding_key: embedding,
                **m,
            }
            for i, t, m, embedding in zip(
                ids, texts, metadatas, embeddings, strict=False
            )
        ]
        operations = [ReplaceOne({"_id": doc["_id"]}, doc, upsert=True) for doc in docs]
        # insert the documents in MongoDB Atlas
        result = self._coll.bulk_write(operations)
        if result.upserted_ids is None:
            raise ValueError("No documents were inserted.")
        return [str(_id) for _id in result.upserted_ids.values()]

    def _run(self, query: str) -> str:
        from bson import json_util

        try:
            query_config = self.query_config or MongoDBVectorSearchConfig()
            limit = query_config.limit
            oversampling_factor = query_config.oversampling_factor
            pre_filter = query_config.pre_filter
            include_embeddings = query_config.include_embeddings
            post_filter_pipeline = query_config.post_filter_pipeline

            # Create the embedding for the query
            query_vector = self._embed_texts([query])[0]

            # Atlas Vector Search, potentially with filter
            stage = {
                "index": self.vector_index_name,
                "path": self.embedding_key,
                "queryVector": query_vector,
                "numCandidates": limit * oversampling_factor,
                "limit": limit,
            }
            if pre_filter:
                stage["filter"] = pre_filter

            pipeline = [
                {"$vectorSearch": stage},
                {"$set": {"score": {"$meta": "vectorSearchScore"}}},
            ]

            # Remove embeddings unless requested
            if not include_embeddings:
                pipeline.append({"$project": {self.embedding_key: 0}})

            # Post-processing
            if post_filter_pipeline is not None:
                pipeline.extend(post_filter_pipeline)

            # Execution
            cursor = self._coll.aggregate(pipeline)  # type: ignore[arg-type]
            docs = []

            # Format
            for doc in cursor:
                docs.append(doc)  # noqa: PERF402
            return json_util.dumps(docs)
        except Exception as e:
            logger.error(f"Error: {e}")
            return ""

    def __del__(self):
        """Cleanup clients on deletion."""
        try:
            if hasattr(self, "_client") and self._client:
                self._client.close()
        except Exception as e:
            logger.error(f"Error: {e}")

        try:
            if hasattr(self, "_openai_client") and self._openai_client:
                self._openai_client.close()
        except Exception as e:
            logger.error(f"Error: {e}")



import json

from crewai_tools.rag.base_loader import BaseLoader, LoaderResult
from crewai_tools.rag.source_content import SourceContent


class JSONLoader(BaseLoader):
    def load(self, source_content: SourceContent, **kwargs) -> LoaderResult:
        source_ref = source_content.source_ref
        content = source_content.source

        if source_content.is_url():
            content = self._load_from_url(source_ref, kwargs)
        elif source_content.path_exists():
            content = self._load_from_file(source_ref)

        return self._parse_json(content, source_ref)

    def _load_from_url(self, url: str, kwargs: dict) -> str:
        import requests

        headers = kwargs.get(
            "headers",
            {
                "Accept": "application/json",
                "User-Agent": "Mozilla/5.0 (compatible; crewai-tools JSONLoader)",
            },
        )

        try:
            response = requests.get(url, headers=headers, timeout=30)
            response.raise_for_status()
            return (
                response.text
                if not self._is_json_response(response)
                else json.dumps(response.json(), indent=2)
            )
        except Exception as e:
            raise ValueError(f"Error fetching JSON from URL {url}: {e!s}") from e

    def _is_json_response(self, response) -> bool:
        try:
            response.json()
            return True
        except ValueError:
            return False

    def _load_from_file(self, path: str) -> str:
        with open(path, "r", encoding="utf-8") as file:
            return file.read()

    def _parse_json(self, content: str, source_ref: str) -> LoaderResult:
        try:
            data = json.loads(content)
            if isinstance(data, dict):
                text = "\n".join(
                    f"{k}: {json.dumps(v, indent=0)}" for k, v in data.items()
                )
            elif isinstance(data, list):
                text = "\n".join(json.dumps(item, indent=0) for item in data)
            else:
                text = json.dumps(data, indent=0)

            metadata = {
                "format": "json",
                "type": type(data).__name__,
                "size": len(data) if isinstance(data, (list, dict)) else 1,
            }
        except json.JSONDecodeError as e:
            text = content
            metadata = {"format": "json", "parse_error": str(e)}

        return LoaderResult(
            content=text,
            source=source_ref,
            metadata=metadata,
            doc_id=self.generate_doc_id(source_ref=source_ref, content=text),
        )



from crewai_tools.rag.base_loader import BaseLoader, LoaderResult
from crewai_tools.rag.source_content import SourceContent


class TextFileLoader(BaseLoader):
    def load(self, source_content: SourceContent, **kwargs) -> LoaderResult:
        source_ref = source_content.source_ref
        if not source_content.path_exists():
            raise FileNotFoundError(
                f"The following file does not exist: {source_content.source}"
            )

        with open(source_content.source, "r", encoding="utf-8") as file:
            content = file.read()

        return LoaderResult(
            content=content,
            source=source_ref,
            doc_id=self.generate_doc_id(source_ref=source_ref, content=content),
        )


class TextLoader(BaseLoader):
    def load(self, source_content: SourceContent, **kwargs) -> LoaderResult:
        return LoaderResult(
            content=source_content.source,
            source=source_content.source_ref,
            doc_id=self.generate_doc_id(content=source_content.source),
        )



import json
import os
from typing import Any

from crewai.tools import BaseTool
from dotenv import load_dotenv
from pydantic import BaseModel, Field

from ..exceptions import BedrockKnowledgeBaseError, BedrockValidationError


# Load environment variables from .env file
load_dotenv()


class BedrockKBRetrieverToolInput(BaseModel):
    """Input schema for BedrockKBRetrieverTool."""

    query: str = Field(
        ..., description="The query to retrieve information from the knowledge base"
    )


class BedrockKBRetrieverTool(BaseTool):
    name: str = "Bedrock Knowledge Base Retriever Tool"
    description: str = (
        "Retrieves information from an Amazon Bedrock Knowledge Base given a query"
    )
    args_schema: type[BaseModel] = BedrockKBRetrieverToolInput
    knowledge_base_id: str = None
    number_of_results: int | None = 5
    retrieval_configuration: dict[str, Any] | None = None
    guardrail_configuration: dict[str, Any] | None = None
    next_token: str | None = None
    package_dependencies: list[str] = Field(default_factory=lambda: ["boto3"])

    def __init__(
        self,
        knowledge_base_id: str | None = None,
        number_of_results: int | None = 5,
        retrieval_configuration: dict[str, Any] | None = None,
        guardrail_configuration: dict[str, Any] | None = None,
        next_token: str | None = None,
        **kwargs,
    ):
        """Initialize the BedrockKBRetrieverTool with knowledge base configuration.

        Args:
            knowledge_base_id (str): The unique identifier of the knowledge base to query
            number_of_results (Optional[int], optional): The maximum number of results to return. Defaults to 5.
            retrieval_configuration (Optional[Dict[str, Any]], optional): Configurations for the knowledge base query and retrieval process. Defaults to None.
            guardrail_configuration (Optional[Dict[str, Any]], optional): Guardrail settings. Defaults to None.
            next_token (Optional[str], optional): Token for retrieving the next batch of results. Defaults to None.
        """
        super().__init__(**kwargs)

        # Get knowledge_base_id from environment variable if not provided
        self.knowledge_base_id = knowledge_base_id or os.getenv("BEDROCK_KB_ID")
        self.number_of_results = number_of_results
        self.guardrail_configuration = guardrail_configuration
        self.next_token = next_token

        # Initialize retrieval_configuration with provided parameters or use the one provided
        if retrieval_configuration is None:
            self.retrieval_configuration = self._build_retrieval_configuration()
        else:
            self.retrieval_configuration = retrieval_configuration

        # Validate parameters
        self._validate_parameters()

        # Update the description to include the knowledge base details
        self.description = f"Retrieves information from Amazon Bedrock Knowledge Base '{self.knowledge_base_id}' given a query"

    def _build_retrieval_configuration(self) -> dict[str, Any]:
        """Build the retrieval configuration based on provided parameters.

        Returns:
            Dict[str, Any]: The constructed retrieval configuration
        """
        vector_search_config = {}

        # Add number of results if provided
        if self.number_of_results is not None:
            vector_search_config["numberOfResults"] = self.number_of_results

        return {"vectorSearchConfiguration": vector_search_config}

    def _validate_parameters(self):
        """Validate the parameters according to AWS API requirements."""
        try:
            # Validate knowledge_base_id
            if not self.knowledge_base_id:
                raise BedrockValidationError("knowledge_base_id cannot be empty")
            if not isinstance(self.knowledge_base_id, str):
                raise BedrockValidationError("knowledge_base_id must be a string")
            if len(self.knowledge_base_id) > 10:
                raise BedrockValidationError(
                    "knowledge_base_id must be 10 characters or less"
                )
            if not all(c.isalnum() for c in self.knowledge_base_id):
                raise BedrockValidationError(
                    "knowledge_base_id must contain only alphanumeric characters"
                )

            # Validate next_token if provided
            if self.next_token:
                if not isinstance(self.next_token, str):
                    raise BedrockValidationError("next_token must be a string")
                if len(self.next_token) < 1 or len(self.next_token) > 2048:
                    raise BedrockValidationError(
                        "next_token must be between 1 and 2048 characters"
                    )
                if " " in self.next_token:
                    raise BedrockValidationError("next_token cannot contain spaces")

            # Validate number_of_results if provided
            if self.number_of_results is not None:
                if not isinstance(self.number_of_results, int):
                    raise BedrockValidationError("number_of_results must be an integer")
                if self.number_of_results < 1:
                    raise BedrockValidationError(
                        "number_of_results must be greater than 0"
                    )

        except BedrockValidationError as e:
            raise BedrockValidationError(f"Parameter validation failed: {e!s}") from e

    def _process_retrieval_result(self, result: dict[str, Any]) -> dict[str, Any]:
        """Process a single retrieval result from Bedrock Knowledge Base.

        Args:
            result (Dict[str, Any]): Raw result from Bedrock Knowledge Base

        Returns:
            Dict[str, Any]: Processed result with standardized format
        """
        # Extract content
        content_obj = result.get("content", {})
        content = content_obj.get("text", "")
        content_type = content_obj.get("type", "text")

        # Extract location information
        location = result.get("location", {})
        location_type = location.get("type", "unknown")
        source_uri = None

        # Map for location types and their URI fields
        location_mapping = {
            "s3Location": {"field": "uri", "type": "S3"},
            "confluenceLocation": {"field": "url", "type": "Confluence"},
            "salesforceLocation": {"field": "url", "type": "Salesforce"},
            "sharePointLocation": {"field": "url", "type": "SharePoint"},
            "webLocation": {"field": "url", "type": "Web"},
            "customDocumentLocation": {"field": "id", "type": "CustomDocument"},
            "kendraDocumentLocation": {"field": "uri", "type": "KendraDocument"},
            "sqlLocation": {"field": "query", "type": "SQL"},
        }

        # Extract the URI based on location type
        for loc_key, config in location_mapping.items():
            if loc_key in location:
                source_uri = location[loc_key].get(config["field"])
                if not location_type or location_type == "unknown":
                    location_type = config["type"]
                break

        # Create result object
        result_object = {
            "content": content,
            "content_type": content_type,
            "source_type": location_type,
            "source_uri": source_uri,
        }

        # Add optional fields if available
        if "score" in result:
            result_object["score"] = result["score"]

        if "metadata" in result:
            result_object["metadata"] = result["metadata"]

        # Handle byte content if present
        if "byteContent" in content_obj:
            result_object["byte_content"] = content_obj["byteContent"]

        # Handle row content if present
        if "row" in content_obj:
            result_object["row_content"] = content_obj["row"]

        return result_object

    def _run(self, query: str) -> str:
        try:
            import boto3
            from botocore.exceptions import ClientError
        except ImportError as e:
            raise ImportError(
                "`boto3` package not found, please run `uv add boto3`"
            ) from e

        try:
            # Initialize the Bedrock Agent Runtime client
            bedrock_agent_runtime = boto3.client(
                "bedrock-agent-runtime",
                region_name=os.getenv(
                    "AWS_REGION", os.getenv("AWS_DEFAULT_REGION", "us-east-1")
                ),
                # AWS SDK will automatically use AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY from environment
            )

            # Prepare the request parameters
            retrieve_params = {
                "knowledgeBaseId": self.knowledge_base_id,
                "retrievalQuery": {"text": query},
            }

            # Add optional parameters if provided
            if self.retrieval_configuration:
                retrieve_params["retrievalConfiguration"] = self.retrieval_configuration

            if self.guardrail_configuration:
                retrieve_params["guardrailConfiguration"] = self.guardrail_configuration

            if self.next_token:
                retrieve_params["nextToken"] = self.next_token

            # Make the retrieve API call
            response = bedrock_agent_runtime.retrieve(**retrieve_params)

            # Process the response
            results = []
            for result in response.get("retrievalResults", []):
                processed_result = self._process_retrieval_result(result)
                results.append(processed_result)

            # Build the response object
            response_object = {}
            if results:
                response_object["results"] = results
            else:
                response_object["message"] = "No results found for the given query."

            if "nextToken" in response:
                response_object["nextToken"] = response["nextToken"]

            if "guardrailAction" in response:
                response_object["guardrailAction"] = response["guardrailAction"]

            # Return the results as a JSON string
            return json.dumps(response_object, indent=2)

        except ClientError as e:
            error_code = "Unknown"
            error_message = str(e)

            # Try to extract error code if available
            if hasattr(e, "response") and "Error" in e.response:
                error_code = e.response["Error"].get("Code", "Unknown")
                error_message = e.response["Error"].get("Message", str(e))

            raise BedrockKnowledgeBaseError(
                f"Error ({error_code}): {error_message}"
            ) from e
        except Exception as e:
            raise BedrockKnowledgeBaseError(f"Unexpected error: {e!s}") from e



from typing import Any

from crewai.tools import BaseTool
from pydantic import BaseModel, Field


try:
    from sqlalchemy import create_engine, text
    from sqlalchemy.orm import sessionmaker

    SQLALCHEMY_AVAILABLE = True
except ImportError:
    SQLALCHEMY_AVAILABLE = False


class NL2SQLToolInput(BaseModel):
    sql_query: str = Field(
        title="SQL Query",
        description="The SQL query to execute.",
    )


class NL2SQLTool(BaseTool):
    name: str = "NL2SQLTool"
    description: str = "Converts natural language to SQL queries and executes them."
    db_uri: str = Field(
        title="Database URI",
        description="The URI of the database to connect to.",
    )
    tables: list = Field(default_factory=list)
    columns: dict = Field(default_factory=dict)
    args_schema: type[BaseModel] = NL2SQLToolInput

    def model_post_init(self, __context: Any) -> None:
        if not SQLALCHEMY_AVAILABLE:
            raise ImportError(
                "sqlalchemy is not installed. Please install it with `pip install crewai-tools[sqlalchemy]`"
            )

        data = {}
        tables = self._fetch_available_tables()

        for table in tables:
            table_columns = self._fetch_all_available_columns(table["table_name"])
            data[f"{table['table_name']}_columns"] = table_columns

        self.tables = tables
        self.columns = data

    def _fetch_available_tables(self):
        return self.execute_sql(
            "SELECT table_name FROM information_schema.tables WHERE table_schema = 'public';"
        )

    def _fetch_all_available_columns(self, table_name: str):
        return self.execute_sql(
            f"SELECT column_name, data_type FROM information_schema.columns WHERE table_name = '{table_name}';"  # noqa: S608
        )

    def _run(self, sql_query: str):
        try:
            data = self.execute_sql(sql_query)
        except Exception as exc:
            data = (
                f"Based on these tables {self.tables} and columns {self.columns}, "
                "you can create SQL queries to retrieve data from the database."
                f"Get the original request {sql_query} and the error {exc} and create the correct SQL query."
            )

        return data

    def execute_sql(self, sql_query: str) -> list | str:
        if not SQLALCHEMY_AVAILABLE:
            raise ImportError(
                "sqlalchemy is not installed. Please install it with `pip install crewai-tools[sqlalchemy]`"
            )

        engine = create_engine(self.db_uri)
        Session = sessionmaker(bind=engine)  # noqa: N806
        session = Session()
        try:
            result = session.execute(text(sql_query))
            session.commit()

            if result.returns_rows:
                columns = result.keys()
                return [
                    dict(zip(columns, row, strict=False)) for row in result.fetchall()
                ]
            return f"Query {sql_query} executed successfully"

        except Exception as e:
            session.rollback()
            raise e

        finally:
            session.close()



import os
import secrets
from typing import Any

from crewai.tools import BaseTool, EnvVar
from openai import OpenAI
from pydantic import BaseModel, Field


class AIMindToolConstants:
    MINDS_API_BASE_URL = "https://mdb.ai/"
    MIND_NAME_PREFIX = "crwai_mind_"
    DATASOURCE_NAME_PREFIX = "crwai_ds_"


class AIMindToolInputSchema(BaseModel):
    """Input for AIMind Tool."""

    query: str = Field(description="Question in natural language to ask the AI-Mind")


class AIMindTool(BaseTool):
    name: str = "AIMind Tool"
    description: str = (
        "A wrapper around [AI-Minds](https://mindsdb.com/minds). "
        "Useful for when you need answers to questions from your data, stored in "
        "data sources including PostgreSQL, MySQL, MariaDB, ClickHouse, Snowflake "
        "and Google BigQuery. "
        "Input should be a question in natural language."
    )
    args_schema: type[BaseModel] = AIMindToolInputSchema
    api_key: str | None = None
    datasources: list[dict[str, Any]] | None = None
    mind_name: str | None = None
    package_dependencies: list[str] = Field(default_factory=lambda: ["minds-sdk"])
    env_vars: list[EnvVar] = Field(
        default_factory=lambda: [
            EnvVar(
                name="MINDS_API_KEY", description="API key for AI-Minds", required=True
            ),
        ]
    )

    def __init__(self, api_key: str | None = None, **kwargs):
        super().__init__(**kwargs)
        self.api_key = api_key or os.getenv("MINDS_API_KEY")
        if not self.api_key:
            raise ValueError(
                "API key must be provided either through constructor or MINDS_API_KEY environment variable"
            )

        try:
            from minds.client import Client  # type: ignore
            from minds.datasources import DatabaseConfig  # type: ignore
        except ImportError as e:
            raise ImportError(
                "`minds_sdk` package not found, please run `pip install minds-sdk`"
            ) from e

        minds_client = Client(api_key=self.api_key)

        # Convert the datasources to DatabaseConfig objects.
        datasources = []
        for datasource in self.datasources:
            config = DatabaseConfig(
                name=f"{AIMindToolConstants.DATASOURCE_NAME_PREFIX}_{secrets.token_hex(5)}",
                engine=datasource["engine"],
                description=datasource["description"],
                connection_data=datasource["connection_data"],
                tables=datasource["tables"],
            )
            datasources.append(config)

        # Generate a random name for the Mind.
        name = f"{AIMindToolConstants.MIND_NAME_PREFIX}_{secrets.token_hex(5)}"

        mind = minds_client.minds.create(
            name=name, datasources=datasources, replace=True
        )

        self.mind_name = mind.name

    def _run(self, query: str):
        # Run the query on the AI-Mind.
        # The Minds API is OpenAI compatible and therefore, the OpenAI client can be used.
        openai_client = OpenAI(
            base_url=AIMindToolConstants.MINDS_API_BASE_URL, api_key=self.api_key
        )

        completion = openai_client.chat.completions.create(
            model=self.mind_name,
            messages=[{"role": "user", "content": query}],
            stream=False,
        )

        return completion.choices[0].message.content



from .snowflake_search_tool import (
    SnowflakeConfig,
    SnowflakeSearchTool,
    SnowflakeSearchToolInput,
)


__all__ = [
    "SnowflakeConfig",
    "SnowflakeSearchTool",
    "SnowflakeSearchToolInput",
]



import os
from unittest.mock import MagicMock, patch

from crewai_tools.tools.generate_crewai_automation_tool.generate_crewai_automation_tool import (
    GenerateCrewaiAutomationTool,
    GenerateCrewaiAutomationToolSchema,
)
import pytest
import requests


@pytest.fixture(autouse=True)
def mock_env():
    with patch.dict(os.environ, {"CREWAI_PERSONAL_ACCESS_TOKEN": "test_token"}):
        os.environ.pop("CREWAI_PLUS_URL", None)
        yield


@pytest.fixture
def tool():
    return GenerateCrewaiAutomationTool()


@pytest.fixture
def custom_url_tool():
    with patch.dict(os.environ, {"CREWAI_PLUS_URL": "https://custom.crewai.com"}):
        return GenerateCrewaiAutomationTool()


def test_default_initialization(tool):
    assert tool.crewai_enterprise_url == "https://app.crewai.com"
    assert tool.personal_access_token == "test_token"
    assert tool.name == "Generate CrewAI Automation"


def test_custom_base_url_from_environment(custom_url_tool):
    assert custom_url_tool.crewai_enterprise_url == "https://custom.crewai.com"


def test_personal_access_token_from_environment(tool):
    assert tool.personal_access_token == "test_token"


def test_valid_prompt_only():
    schema = GenerateCrewaiAutomationToolSchema(
        prompt="Create a web scraping automation"
    )
    assert schema.prompt == "Create a web scraping automation"
    assert schema.organization_id is None


def test_valid_prompt_with_organization_id():
    schema = GenerateCrewaiAutomationToolSchema(
        prompt="Create automation", organization_id="org-123"
    )
    assert schema.prompt == "Create automation"
    assert schema.organization_id == "org-123"


def test_empty_prompt_validation():
    schema = GenerateCrewaiAutomationToolSchema(prompt="")
    assert schema.prompt == ""
    assert schema.organization_id is None


@patch("requests.post")
def test_successful_generation_without_org_id(mock_post, tool):
    mock_response = MagicMock()
    mock_response.json.return_value = {
        "url": "https://app.crewai.com/studio/project-123"
    }
    mock_post.return_value = mock_response

    result = tool.run(prompt="Create automation")

    assert (
        result
        == "Generated CrewAI Studio project URL: https://app.crewai.com/studio/project-123"
    )
    mock_post.assert_called_once_with(
        "https://app.crewai.com/crewai_plus/api/v1/studio",
        headers={
            "Authorization": "Bearer test_token",
            "Content-Type": "application/json",
            "Accept": "application/json",
        },
        json={"prompt": "Create automation"},
    )


@patch("requests.post")
def test_successful_generation_with_org_id(mock_post, tool):
    mock_response = MagicMock()
    mock_response.json.return_value = {
        "url": "https://app.crewai.com/studio/project-456"
    }
    mock_post.return_value = mock_response

    result = tool.run(prompt="Create automation", organization_id="org-456")

    assert (
        result
        == "Generated CrewAI Studio project URL: https://app.crewai.com/studio/project-456"
    )
    mock_post.assert_called_once_with(
        "https://app.crewai.com/crewai_plus/api/v1/studio",
        headers={
            "Authorization": "Bearer test_token",
            "Content-Type": "application/json",
            "Accept": "application/json",
            "X-Crewai-Organization-Id": "org-456",
        },
        json={"prompt": "Create automation"},
    )


@patch("requests.post")
def test_custom_base_url_usage(mock_post, custom_url_tool):
    mock_response = MagicMock()
    mock_response.json.return_value = {
        "url": "https://custom.crewai.com/studio/project-789"
    }
    mock_post.return_value = mock_response

    custom_url_tool.run(prompt="Create automation")

    mock_post.assert_called_once_with(
        "https://custom.crewai.com/crewai_plus/api/v1/studio",
        headers={
            "Authorization": "Bearer test_token",
            "Content-Type": "application/json",
            "Accept": "application/json",
        },
        json={"prompt": "Create automation"},
    )


@patch("requests.post")
def test_api_error_response_handling(mock_post, tool):
    mock_post.return_value.raise_for_status.side_effect = requests.HTTPError(
        "400 Bad Request"
    )

    with pytest.raises(requests.HTTPError):
        tool.run(prompt="Create automation")


@patch("requests.post")
def test_network_error_handling(mock_post, tool):
    mock_post.side_effect = requests.ConnectionError("Network unreachable")

    with pytest.raises(requests.ConnectionError):
        tool.run(prompt="Create automation")


@patch("requests.post")
def test_api_response_missing_url(mock_post, tool):
    mock_response = MagicMock()
    mock_response.json.return_value = {"status": "success"}
    mock_post.return_value = mock_response

    result = tool.run(prompt="Create automation")

    assert result == "Generated CrewAI Studio project URL: None"


def test_authorization_header_construction(tool):
    headers = tool._get_headers()

    assert headers["Authorization"] == "Bearer test_token"
    assert headers["Content-Type"] == "application/json"
    assert headers["Accept"] == "application/json"
    assert "X-Crewai-Organization-Id" not in headers


def test_authorization_header_with_org_id(tool):
    headers = tool._get_headers(organization_id="org-123")

    assert headers["Authorization"] == "Bearer test_token"
    assert headers["X-Crewai-Organization-Id"] == "org-123"


def test_missing_personal_access_token():
    with patch.dict(os.environ, {}, clear=True):
        tool = GenerateCrewaiAutomationTool()
        assert tool.personal_access_token is None



import logging
import os

from crewai.tools import BaseTool
from pydantic import Field, create_model
import requests


ACTIONS_URL = "https://actions.zapier.com/api/v2/ai-actions"

logger = logging.getLogger(__name__)


class ZapierActionTool(BaseTool):
    """A tool that wraps a Zapier action."""

    name: str = Field(description="Tool name")
    description: str = Field(description="Tool description")
    action_id: str = Field(description="Zapier action ID")
    api_key: str = Field(description="Zapier API key")

    def _run(self, **kwargs) -> str:
        """Execute the Zapier action."""
        headers = {"x-api-key": self.api_key, "Content-Type": "application/json"}

        instructions = kwargs.pop(
            "instructions", "Execute this action with the provided parameters"
        )

        if not kwargs:
            action_params = {"instructions": instructions, "params": {}}
        else:
            formatted_params = {}
            for key, value in kwargs.items():
                formatted_params[key] = {
                    "value": value,
                    "mode": "guess",
                }
            action_params = {"instructions": instructions, "params": formatted_params}

        execute_url = f"{ACTIONS_URL}/{self.action_id}/execute/"
        response = requests.request(
            "POST",
            execute_url,
            headers=headers,
            json=action_params,
            timeout=30,
        )

        response.raise_for_status()

        return response.json()


class ZapierActionsAdapter:
    """Adapter for Zapier Actions."""

    api_key: str

    def __init__(self, api_key: str | None = None):
        self.api_key = api_key or os.getenv("ZAPIER_API_KEY")
        if not self.api_key:
            logger.error("Zapier Actions API key is required")
            raise ValueError("Zapier Actions API key is required")

    def get_zapier_actions(self):
        headers = {
            "x-api-key": self.api_key,
        }
        response = requests.request(
            "GET",
            ACTIONS_URL,
            headers=headers,
            timeout=30,
        )
        response.raise_for_status()

        return response.json()

    def tools(self) -> list[BaseTool]:
        """Convert Zapier actions to BaseTool instances."""
        actions_response = self.get_zapier_actions()
        tools = []

        for action in actions_response.get("results", []):
            tool_name = (
                action["meta"]["action_label"]
                .replace(" ", "_")
                .replace(":", "")
                .lower()
            )

            params = action.get("params", {})
            args_fields = {}

            args_fields["instructions"] = (
                str,
                Field(description="Instructions for how to execute this action"),
            )

            for param_name, param_info in params.items():
                field_type = (
                    str  # Default to string, could be enhanced based on param_info
                )
                field_description = (
                    param_info.get("description", "")
                    if isinstance(param_info, dict)
                    else ""
                )
                args_fields[param_name] = (
                    field_type,
                    Field(description=field_description),
                )

            args_schema = create_model(f"{tool_name.title()}Schema", **args_fields)

            tool = ZapierActionTool(
                name=tool_name,
                description=action["description"],
                action_id=action["id"],
                api_key=self.api_key,
                args_schema=args_schema,
            )
            tools.append(tool)

        return tools



import re
import time
from typing import Any
from urllib.parse import urlparse

from crewai.tools import BaseTool
from pydantic import BaseModel, Field, field_validator


class FixedSeleniumScrapingToolSchema(BaseModel):
    """Input for SeleniumScrapingTool."""


class SeleniumScrapingToolSchema(FixedSeleniumScrapingToolSchema):
    """Input for SeleniumScrapingTool."""

    website_url: str = Field(
        ...,
        description="Mandatory website url to read the file. Must start with http:// or https://",
    )
    css_element: str = Field(
        ...,
        description="Mandatory css reference for element to scrape from the website",
    )

    @field_validator("website_url")
    @classmethod
    def validate_website_url(cls, v):
        if not v:
            raise ValueError("Website URL cannot be empty")

        if len(v) > 2048:  # Common maximum URL length
            raise ValueError("URL is too long (max 2048 characters)")

        if not re.match(r"^https?://", v):
            raise ValueError("URL must start with http:// or https://")

        try:
            result = urlparse(v)
            if not all([result.scheme, result.netloc]):
                raise ValueError("Invalid URL format")
        except Exception as e:
            raise ValueError(f"Invalid URL: {e!s}") from e

        if re.search(r"\s", v):
            raise ValueError("URL cannot contain whitespace")

        return v


class SeleniumScrapingTool(BaseTool):
    name: str = "Read a website content"
    description: str = "A tool that can be used to read a website content."
    args_schema: type[BaseModel] = SeleniumScrapingToolSchema
    website_url: str | None = None
    driver: Any | None = None
    cookie: dict | None = None
    wait_time: int | None = 3
    css_element: str | None = None
    return_html: bool | None = False
    _by: Any | None = None
    package_dependencies: list[str] = Field(
        default_factory=lambda: ["selenium", "webdriver-manager"]
    )

    def __init__(
        self,
        website_url: str | None = None,
        cookie: dict | None = None,
        css_element: str | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        try:
            from selenium import webdriver  # type: ignore[import-not-found]
            from selenium.webdriver.chrome.options import (  # type: ignore[import-not-found]
                Options,
            )
            from selenium.webdriver.common.by import (  # type: ignore[import-not-found]
                By,
            )
        except ImportError:
            import click

            if click.confirm(
                "You are missing the 'selenium' and 'webdriver-manager' packages. Would you like to install it?"
            ):
                import subprocess

                subprocess.run(
                    ["uv", "pip", "install", "selenium", "webdriver-manager"],  # noqa: S607
                    check=True,
                )
                from selenium import webdriver  # type: ignore[import-not-found]
                from selenium.webdriver.chrome.options import (  # type: ignore[import-not-found]
                    Options,
                )
                from selenium.webdriver.common.by import (  # type: ignore[import-not-found]
                    By,
                )
            else:
                raise ImportError(
                    "`selenium` and `webdriver-manager` package not found, please run `uv add selenium webdriver-manager`"
                ) from None

        if "driver" not in kwargs:
            if "options" not in kwargs:
                options: Options = Options()
                options.add_argument("--headless")
            else:
                options = kwargs["options"]
            self.driver = webdriver.Chrome(options=options)
        else:
            self.driver = kwargs["driver"]

        self._by = By
        if cookie is not None:
            self.cookie = cookie

        if css_element is not None:
            self.css_element = css_element

        if website_url is not None:
            self.website_url = website_url
            self.description = (
                f"A tool that can be used to read {website_url}'s content."
            )
            self.args_schema = FixedSeleniumScrapingToolSchema

        self._generate_description()

    def _run(
        self,
        **kwargs: Any,
    ) -> Any:
        website_url = kwargs.get("website_url", self.website_url)
        css_element = kwargs.get("css_element", self.css_element)
        return_html = kwargs.get("return_html", self.return_html)
        try:
            self._make_request(website_url, self.cookie, self.wait_time)
            content = self._get_content(css_element, return_html)
            return "\n".join(content)
        except Exception as e:
            return f"Error scraping website: {e!s}"
        finally:
            if self.driver is not None:
                self.driver.close()

    def _get_content(self, css_element, return_html):
        content = []

        if self._is_css_element_empty(css_element):
            content.append(self._get_body_content(return_html))
        else:
            content.extend(self._get_elements_content(css_element, return_html))

        return content

    def _is_css_element_empty(self, css_element):
        return css_element is None or css_element.strip() == ""

    def _get_body_content(self, return_html):
        body_element = self.driver.find_element(self._by.TAG_NAME, "body")

        return (
            body_element.get_attribute("outerHTML")
            if return_html
            else body_element.text
        )

    def _get_elements_content(self, css_element, return_html):
        elements_content = []

        for element in self.driver.find_elements(self._by.CSS_SELECTOR, css_element):
            elements_content.append(  # noqa: PERF401
                element.get_attribute("outerHTML") if return_html else element.text
            )

        return elements_content

    def _make_request(self, url, cookie, wait_time):
        if not url:
            raise ValueError("URL cannot be empty")

        # Validate URL format
        if not re.match(r"^https?://", url):
            raise ValueError("URL must start with http:// or https://")

        self.driver.get(url)
        time.sleep(wait_time)
        if cookie:
            self.driver.add_cookie(cookie)
            time.sleep(wait_time)
            self.driver.get(url)
            time.sleep(wait_time)

    def close(self):
        self.driver.close()



import hashlib
import os
import tempfile

from crewai_tools.rag.base_loader import LoaderResult
from crewai_tools.rag.loaders.text_loader import TextFileLoader, TextLoader
from crewai_tools.rag.source_content import SourceContent
import pytest


def write_temp_file(content, suffix=".txt", encoding="utf-8"):
    with tempfile.NamedTemporaryFile(
        mode="w", suffix=suffix, delete=False, encoding=encoding
    ) as f:
        f.write(content)
        return f.name


def cleanup_temp_file(path):
    try:
        os.unlink(path)
    except FileNotFoundError:
        pass


class TestTextFileLoader:
    def test_basic_text_file(self):
        content = "This is test content\nWith multiple lines\nAnd more text"
        path = write_temp_file(content)
        try:
            result = TextFileLoader().load(SourceContent(path))
            assert isinstance(result, LoaderResult)
            assert result.content == content
            assert result.source == path
            assert result.doc_id
            assert result.metadata in (None, {})
        finally:
            cleanup_temp_file(path)

    def test_empty_file(self):
        path = write_temp_file("")
        try:
            result = TextFileLoader().load(SourceContent(path))
            assert result.content == ""
        finally:
            cleanup_temp_file(path)

    def test_unicode_content(self):
        content = "Hello 世界 🌍 émojis 🎉 åäö"
        path = write_temp_file(content)
        try:
            result = TextFileLoader().load(SourceContent(path))
            assert content in result.content
        finally:
            cleanup_temp_file(path)

    def test_large_file(self):
        content = "\n".join(f"Line {i}" for i in range(100))
        path = write_temp_file(content)
        try:
            result = TextFileLoader().load(SourceContent(path))
            assert "Line 0" in result.content
            assert "Line 99" in result.content
            assert result.content.count("\n") == 99
        finally:
            cleanup_temp_file(path)

    def test_missing_file(self):
        with pytest.raises(FileNotFoundError):
            TextFileLoader().load(SourceContent("/nonexistent/path.txt"))

    def test_permission_denied(self):
        path = write_temp_file("Some content")
        os.chmod(path, 0o000)
        try:
            with pytest.raises(PermissionError):
                TextFileLoader().load(SourceContent(path))
        finally:
            os.chmod(path, 0o644)
            cleanup_temp_file(path)

    def test_doc_id_consistency(self):
        content = "Consistent content"
        path = write_temp_file(content)
        try:
            loader = TextFileLoader()
            result1 = loader.load(SourceContent(path))
            result2 = loader.load(SourceContent(path))
            expected_id = hashlib.sha256((path + content).encode("utf-8")).hexdigest()
            assert result1.doc_id == result2.doc_id == expected_id
        finally:
            cleanup_temp_file(path)

    def test_various_extensions(self):
        content = "Same content"
        for ext in [".txt", ".md", ".log", ".json"]:
            path = write_temp_file(content, suffix=ext)
            try:
                result = TextFileLoader().load(SourceContent(path))
                assert result.content == content
            finally:
                cleanup_temp_file(path)


class TestTextLoader:
    def test_basic_text(self):
        content = "Raw text"
        result = TextLoader().load(SourceContent(content))
        expected_hash = hashlib.sha256(content.encode("utf-8")).hexdigest()
        assert result.content == content
        assert result.source == expected_hash
        assert result.doc_id == expected_hash

    def test_multiline_text(self):
        content = "Line 1\nLine 2\nLine 3"
        result = TextLoader().load(SourceContent(content))
        assert "Line 2" in result.content

    def test_empty_text(self):
        result = TextLoader().load(SourceContent(""))
        assert result.content == ""
        assert result.source == hashlib.sha256("".encode("utf-8")).hexdigest()

    def test_unicode_text(self):
        content = "世界 🌍 émojis 🎉 åäö"
        result = TextLoader().load(SourceContent(content))
        assert content in result.content

    def test_special_characters(self):
        content = "!@#$$%^&*()_+-=~`{}[]\\|;:'\",.<>/?"
        result = TextLoader().load(SourceContent(content))
        assert result.content == content

    def test_doc_id_uniqueness(self):
        result1 = TextLoader().load(SourceContent("A"))
        result2 = TextLoader().load(SourceContent("B"))
        assert result1.doc_id != result2.doc_id

    def test_whitespace_text(self):
        content = "   \n\t   "
        result = TextLoader().load(SourceContent(content))
        assert result.content == content

    def test_long_text(self):
        content = "A" * 10000
        result = TextLoader().load(SourceContent(content))
        assert len(result.content) == 10000


class TestTextLoadersIntegration:
    def test_consistency_between_loaders(self):
        content = "Consistent content"
        text_result = TextLoader().load(SourceContent(content))
        file_path = write_temp_file(content)
        try:
            file_result = TextFileLoader().load(SourceContent(file_path))

            assert text_result.content == file_result.content
            assert text_result.source != file_result.source
            assert text_result.doc_id != file_result.doc_id
        finally:
            cleanup_temp_file(file_path)



from crewai.tools import BaseTool
from pydantic import BaseModel, Field


class ContextualAIParseSchema(BaseModel):
    """Schema for contextual parse tool."""

    file_path: str = Field(..., description="Path to the document to parse")
    parse_mode: str = Field(default="standard", description="Parsing mode")
    figure_caption_mode: str = Field(
        default="concise", description="Figure caption mode"
    )
    enable_document_hierarchy: bool = Field(
        default=True, description="Enable document hierarchy"
    )
    page_range: str | None = Field(
        default=None, description="Page range to parse (e.g., '0-5')"
    )
    output_types: list[str] = Field(
        default=["markdown-per-page"], description="List of output types"
    )


class ContextualAIParseTool(BaseTool):
    """Tool to parse documents using Contextual AI's parser."""

    name: str = "Contextual AI Document Parser"
    description: str = "Parse documents using Contextual AI's advanced document parser"
    args_schema: type[BaseModel] = ContextualAIParseSchema

    api_key: str
    package_dependencies: list[str] = Field(
        default_factory=lambda: ["contextual-client"]
    )

    def _run(
        self,
        file_path: str,
        parse_mode: str = "standard",
        figure_caption_mode: str = "concise",
        enable_document_hierarchy: bool = True,
        page_range: str | None = None,
        output_types: list[str] | None = None,
    ) -> str:
        """Parse a document using Contextual AI's parser."""
        if output_types is None:
            output_types = ["markdown-per-page"]
        try:
            import json
            import os
            from time import sleep

            import requests

            if not os.path.exists(file_path):
                raise FileNotFoundError(f"Document not found: {file_path}")

            base_url = "https://api.contextual.ai/v1"
            headers = {
                "accept": "application/json",
                "authorization": f"Bearer {self.api_key}",
            }

            # Submit parse job
            url = f"{base_url}/parse"
            config = {
                "parse_mode": parse_mode,
                "figure_caption_mode": figure_caption_mode,
                "enable_document_hierarchy": enable_document_hierarchy,
            }

            if page_range:
                config["page_range"] = page_range

            with open(file_path, "rb") as fp:
                file = {"raw_file": fp}
                result = requests.post(
                    url, headers=headers, data=config, files=file, timeout=30
                )
                response = json.loads(result.text)
                job_id = response["job_id"]

            # Monitor job status
            status_url = f"{base_url}/parse/jobs/{job_id}/status"
            while True:
                result = requests.get(status_url, headers=headers, timeout=30)
                parse_response = json.loads(result.text)["status"]

                if parse_response == "completed":
                    break
                if parse_response == "failed":
                    raise RuntimeError("Document parsing failed")

                sleep(5)

            # Get parse results
            results_url = f"{base_url}/parse/jobs/{job_id}/results"
            result = requests.get(
                results_url,
                headers=headers,
                params={"output_types": ",".join(output_types)},
                timeout=30,
            )

            return json.dumps(json.loads(result.text), indent=2)

        except Exception as e:
            return f"Failed to parse document: {e!s}"



import os
from typing import Any, Literal

from crewai.tools import BaseTool, EnvVar
from pydantic import BaseModel, Field


class HyperbrowserLoadToolSchema(BaseModel):
    url: str = Field(description="Website URL")
    operation: Literal["scrape", "crawl"] = Field(
        description="Operation to perform on the website. Either 'scrape' or 'crawl'"
    )
    params: dict | None = Field(
        description="Optional params for scrape or crawl. For more information on the supported params, visit https://docs.hyperbrowser.ai/reference/sdks/python/scrape#start-scrape-job-and-wait or https://docs.hyperbrowser.ai/reference/sdks/python/crawl#start-crawl-job-and-wait"
    )


class HyperbrowserLoadTool(BaseTool):
    """HyperbrowserLoadTool.

    Scrape or crawl web pages and load the contents with optional parameters for configuring content extraction.
    Requires the `hyperbrowser` package.
    Get your API Key from https://app.hyperbrowser.ai/

    Args:
        api_key: The Hyperbrowser API key, can be set as an environment variable `HYPERBROWSER_API_KEY` or passed directly
    """

    name: str = "Hyperbrowser web load tool"
    description: str = "Scrape or crawl a website using Hyperbrowser and return the contents in properly formatted markdown or html"
    args_schema: type[BaseModel] = HyperbrowserLoadToolSchema
    api_key: str | None = None
    hyperbrowser: Any | None = None
    package_dependencies: list[str] = Field(default_factory=lambda: ["hyperbrowser"])
    env_vars: list[EnvVar] = Field(
        default_factory=lambda: [
            EnvVar(
                name="HYPERBROWSER_API_KEY",
                description="API key for Hyperbrowser services",
                required=False,
            ),
        ]
    )

    def __init__(self, api_key: str | None = None, **kwargs):
        super().__init__(**kwargs)
        self.api_key = api_key or os.getenv("HYPERBROWSER_API_KEY")
        if not api_key:
            raise ValueError(
                "`api_key` is required, please set the `HYPERBROWSER_API_KEY` environment variable or pass it directly"
            )

        try:
            from hyperbrowser import Hyperbrowser
        except ImportError as e:
            raise ImportError(
                "`hyperbrowser` package not found, please run `pip install hyperbrowser`"
            ) from e

        if not self.api_key:
            raise ValueError(
                "HYPERBROWSER_API_KEY is not set. Please provide it either via the constructor with the `api_key` argument or by setting the HYPERBROWSER_API_KEY environment variable."
            )

        self.hyperbrowser = Hyperbrowser(api_key=self.api_key)

    def _prepare_params(self, params: dict) -> dict:
        """Prepare session and scrape options parameters."""
        try:
            from hyperbrowser.models.scrape import ScrapeOptions
            from hyperbrowser.models.session import CreateSessionParams
        except ImportError as e:
            raise ImportError(
                "`hyperbrowser` package not found, please run `pip install hyperbrowser`"
            ) from e

        if "scrape_options" in params:
            if "formats" in params["scrape_options"]:
                formats = params["scrape_options"]["formats"]
                if not all(fmt in ["markdown", "html"] for fmt in formats):
                    raise ValueError("formats can only contain 'markdown' or 'html'")

        if "session_options" in params:
            params["session_options"] = CreateSessionParams(**params["session_options"])
        if "scrape_options" in params:
            params["scrape_options"] = ScrapeOptions(**params["scrape_options"])
        return params

    def _extract_content(self, data: Any | None):
        """Extract content from response data."""
        content = ""
        if data:
            content = data.markdown or data.html or ""
        return content

    def _run(
        self,
        url: str,
        operation: Literal["scrape", "crawl"] = "scrape",
        params: dict | None = None,
    ):
        if params is None:
            params = {}
        try:
            from hyperbrowser.models.crawl import StartCrawlJobParams
            from hyperbrowser.models.scrape import StartScrapeJobParams
        except ImportError as e:
            raise ImportError(
                "`hyperbrowser` package not found, please run `pip install hyperbrowser`"
            ) from e

        params = self._prepare_params(params)

        if operation == "scrape":
            scrape_params = StartScrapeJobParams(url=url, **params)
            scrape_resp = self.hyperbrowser.scrape.start_and_wait(scrape_params)
            return self._extract_content(scrape_resp.data)
        crawl_params = StartCrawlJobParams(url=url, **params)
        crawl_resp = self.hyperbrowser.crawl.start_and_wait(crawl_params)
        content = ""
        if crawl_resp.data:
            for page in crawl_resp.data:
                page_content = self._extract_content(page)
                if page_content:
                    content += (
                        f"\n{'-' * 50}\nUrl: {page.url}\nContent:\n{page_content}\n"
                    )
        return content
